@@ -3,6 +3,7 @@
 #define FISH_BUILTIN_H
 
 #include <stddef.h>
+
 #include <vector>
 
 #include "common.h"
@@ -56,13 +57,11 @@ enum { COMMAND_NOT_BUILTIN, BUILTIN_REGULAR, BUILTIN_FUNCTION };
 #define BUILTIN_ERR_ARG_COUNT1 _(L"%ls: expected %d args, got %d\n")
 #define BUILTIN_ERR_ARG_COUNT2 _(L"%ls: %ls expected %d args, got %d\n")
 
-/// Error message for invalid character in variable name.
-#define BUILTIN_ERR_VARCHAR                                                                 \
-    _(L"%ls: Invalid character '%lc' in variable name. Only alphanumerical characters and " \
-      L"underscores are valid in a variable name.\n")
+/// Error message for invalid variable name.
+#define BUILTIN_ERR_VARNAME _(L"%ls: Variable name '%ls' is not valid. See `help identifiers`.\n")
 
-/// Error message for invalid (empty) variable name.
-#define BUILTIN_ERR_VARNAME_ZERO _(L"%ls: Variable name can not be the empty string\n")
+/// Error message for invalid bind mode name.
+#define BUILTIN_ERR_BIND_MODE _(L"%ls: mode name '%ls' is not valid. See `help identifiers`.\n")
 
 /// Error message when too many arguments are supplied to a builtin.
 #define BUILTIN_ERR_TOO_MANY_ARGUMENTS _(L"%ls: Too many arguments\n")
