@@ -28,13 +28,12 @@ enum {
 /// Calculates the cpu usage (in percent) of the specified job.
 static int cpu_use(const job_t *j) {
     double u = 0;
-    process_t *p;
 
-    for (p = j->first_process; p; p = p->next) {
+    for (const process_ptr_t &p : j->processes) {
         struct timeval t;
         int jiffies;
         gettimeofday(&t, 0);
-        jiffies = proc_get_jiffies(p);
+        jiffies = proc_get_jiffies(p.get());
 
         double t1 = 1000000.0 * p->last_time.tv_sec + p->last_time.tv_usec;
         double t2 = 1000000.0 * t.tv_sec + t.tv_usec;
@@ -48,7 +47,6 @@ static int cpu_use(const job_t *j) {
 
 /// Print information about the specified job.
 static void builtin_jobs_print(const job_t *j, int mode, int header, io_streams_t &streams) {
-    process_t *p;
     switch (mode) {
         case JOBS_DEFAULT: {
             if (header) {
@@ -85,7 +83,7 @@ static void builtin_jobs_print(const job_t *j, int mode, int header, io_streams_
                 streams.out.append(_(L"Process\n"));
             }
 
-            for (p = j->first_process; p; p = p->next) {
+            for (const process_ptr_t &p : j->processes) {
                 streams.out.append_format(L"%d\n", p->pid);
             }
             break;
@@ -96,11 +94,15 @@ static void builtin_jobs_print(const job_t *j, int mode, int header, io_streams_
                 streams.out.append(_(L"Command\n"));
             }
 
-            for (p = j->first_process; p; p = p->next) {
+            for (const process_ptr_t &p : j->processes) {
                 streams.out.append_format(L"%ls\n", p->argv0());
             }
             break;
         }
+        default: {
+            DIE("unexpected mode");
+            break;
+        }
     }
 }
 
@@ -153,11 +155,15 @@ int builtin_jobs(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             }
             case 'h': {
                 builtin_print_help(parser, streams, argv[0], streams.out);
-                return 0;
+                return STATUS_CMD_OK;
             }
             case '?': {
                 builtin_unknown_option(parser, streams, argv[0], argv[w.woptind - 1]);
-                return 1;
+                return STATUS_INVALID_ARGS;
+            }
+            default: {
+                DIE("unexpected opt");
+                break;
             }
         }
     }
@@ -169,7 +175,7 @@ int builtin_jobs(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         while ((j = jobs.next())) {
             if ((j->flags & JOB_CONSTRUCTED) && !job_is_completed(j)) {
                 builtin_jobs_print(j, mode, !streams.out_is_redirected, streams);
-                return 0;
+                return STATUS_CMD_ERROR;
             }
         }
 
@@ -178,13 +184,10 @@ int builtin_jobs(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             int i;
 
             for (i = w.woptind; i < argc; i++) {
-                int pid;
-                wchar_t *end;
-                errno = 0;
-                pid = fish_wcstoi(argv[i], &end, 10);
-                if (errno || *end) {
+                int pid = fish_wcstoi(argv[i]);
+                if (errno || pid < 0) {
                     streams.err.append_format(_(L"%ls: '%ls' is not a job\n"), argv[0], argv[i]);
-                    return 1;
+                    return STATUS_INVALID_ARGS;
                 }
 
                 const job_t *j = job_get_from_pid(pid);
@@ -194,7 +197,7 @@ int builtin_jobs(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                     found = 1;
                 } else {
                     streams.err.append_format(_(L"%ls: No suitable job: %d\n"), argv[0], pid);
-                    return 1;
+                    return STATUS_CMD_ERROR;
                 }
             }
         } else {
@@ -215,8 +218,8 @@ int builtin_jobs(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         if (!streams.out_is_redirected) {
             streams.out.append_format(_(L"%ls: There are no jobs\n"), argv[0]);
         }
-        return 1;
+        return STATUS_CMD_ERROR;
     }
 
-    return 0;
+    return STATUS_CMD_OK;
 }
