@@ -54,7 +54,9 @@
 #include <limits.h>
 #include <locale.h>
 #include <stdarg.h>
-#include <stdio.h>
+#include <stddef.h>
+#include <stdint.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <wchar.h>
@@ -63,7 +65,6 @@
 #include "builtin.h"
 #include "common.h"
 #include "io.h"
-#include "proc.h"
 #include "wutil.h"  // IWYU pragma: keep
 
 class parser_t;
@@ -207,7 +208,7 @@ void builtin_printf_state_t::fatal_error(const wchar_t *fmt, ...) {
     streams.err.append(errstr);
     if (!string_suffixes_string(L"\n", errstr)) streams.err.push_back(L'\n');
 
-    this->exit_code = STATUS_BUILTIN_ERROR;
+    this->exit_code = STATUS_CMD_ERROR;
     this->early_exit = true;
 }
 
@@ -238,7 +239,11 @@ void builtin_printf_state_t::append_format_output(const wchar_t *fmt, ...) {
 
 void builtin_printf_state_t::verify_numeric(const wchar_t *s, const wchar_t *end, int errcode) {
     if (errcode != 0) {
-        this->fatal_error(L"%ls: %s", s, strerror(errcode));
+        if (errcode == ERANGE) {
+            this->fatal_error(L"%ls: %ls", s, _(L"Number out of range"));
+        } else {
+            this->fatal_error(L"%ls: %s", s, strerror(errcode));
+        }
     } else if (*end) {
         if (s == end)
             this->fatal_error(_(L"%ls: expected a numeric value"), s);
@@ -312,7 +317,7 @@ void builtin_printf_state_t::print_esc_char(wchar_t c) {
             break;
         }
         case L'e': {  // escape
-            this->append_output(L'\x1B');
+            this->append_output(L'\e');
             break;
         }
         case L'f': {  // form feed
@@ -435,6 +440,8 @@ void builtin_printf_state_t::print_direc(const wchar_t *start, size_t length, wc
     // Create a copy of the % directive, with an intmax_t-wide width modifier substituted for any
     // existing integer length modifier.
     switch (conversion) {
+        case L'x':
+        case L'X':
         case L'd':
         case L'i':
         case L'u': {
@@ -508,39 +515,48 @@ void builtin_printf_state_t::print_direc(const wchar_t *start, size_t length, wc
         case L'G': {
             long double arg = string_to_scalar_type<long double>(argument, this);
             if (!have_field_width) {
-                if (!have_precision)
+                if (!have_precision) {
                     this->append_format_output(fmt.c_str(), arg);
-                else
+                } else {
                     this->append_format_output(fmt.c_str(), precision, arg);
+                }
             } else {
-                if (!have_precision)
+                if (!have_precision) {
                     this->append_format_output(fmt.c_str(), field_width, arg);
-                else
+                } else {
                     this->append_format_output(fmt.c_str(), field_width, precision, arg);
+                }
             }
             break;
         }
         case L'c': {
-            if (!have_field_width)
+            if (!have_field_width) {
                 this->append_format_output(fmt.c_str(), *argument);
-            else
+            } else {
                 this->append_format_output(fmt.c_str(), field_width, *argument);
+            }
             break;
         }
         case L's': {
             if (!have_field_width) {
                 if (!have_precision) {
                     this->append_format_output(fmt.c_str(), argument);
-                } else
+                } else {
                     this->append_format_output(fmt.c_str(), precision, argument);
+                }
             } else {
-                if (!have_precision)
+                if (!have_precision) {
                     this->append_format_output(fmt.c_str(), field_width, argument);
-                else
+                } else {
                     this->append_format_output(fmt.c_str(), field_width, precision, argument);
+                }
             }
             break;
         }
+        default: {
+            DIE("unexpected opt");
+            break;
+        }
     }
 }
 
@@ -588,8 +604,7 @@ int builtin_printf_state_t::print_formatted(const wchar_t *format, int argc, wch
                 }
 
                 modify_allowed_format_specifiers(ok, "aAcdeEfFgGiosuxX", true);
-
-                for (;; f++, direc_length++) {
+                for (bool continue_looking_for_flags = true; continue_looking_for_flags;) {
                     switch (*f) {
                         case L'I':
                         case L'\'': {
@@ -609,10 +624,16 @@ int builtin_printf_state_t::print_formatted(const wchar_t *format, int argc, wch
                             modify_allowed_format_specifiers(ok, "cs", false);
                             break;
                         }
-                        default: { goto no_more_flag_characters; }
+                        default: {
+                            continue_looking_for_flags = false;
+                            break;
+                        }
+                    }
+                    if (continue_looking_for_flags) {
+                        f++;
+                        direc_length++;
                     }
                 }
-            no_more_flag_characters:;
 
                 if (*f == L'*') {
                     ++f;
@@ -687,7 +708,10 @@ int builtin_printf_state_t::print_formatted(const wchar_t *format, int argc, wch
                 f += print_esc(f, false);
                 break;
             }
-            default: { this->append_output(*f); }
+            default: {
+                this->append_output(*f);
+                break;
+            }
         }
     }
     return save_argc - argc;
@@ -704,7 +728,7 @@ int builtin_printf(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
 
     if (argc <= 1) {
         state.fatal_error(_(L"printf: not enough arguments"));
-        return STATUS_BUILTIN_ERROR;
+        return STATUS_INVALID_ARGS;
     }
 
     format = argv[1];
