@@ -2,10 +2,14 @@
 #include "config.h"  // IWYU pragma: keep
 
 #include <errno.h>
+#include <stddef.h>
 #include <stdlib.h>
+#include <string.h>
 #include <sys/stat.h>
+#include <unistd.h>
 #include <wchar.h>
 #include <wctype.h>
+
 #include <algorithm>
 #include <iterator>
 #include <memory>
@@ -36,7 +40,8 @@ class parser_t;
     L"%ls: The number of variable indexes does not match the number of values\n"
 
 // Test if the specified variable should be subject to path validation.
-static int is_path_variable(const wchar_t *env) { return contains(env, L"PATH", L"CDPATH"); }
+static const wcstring_list_t path_variables({L"PATH", L"CDPATH"});
+static int is_path_variable(const wchar_t *env) { return contains(path_variables, env); }
 
 /// Call env_set. If this is a path variable, e.g. PATH, validate the elements. On error, print a
 /// description of the problem to stderr.
@@ -67,7 +72,7 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
 
         for (i = 0; i < val.size(); i++) {
             const wcstring &dir = val.at(i);
-            if (!string_prefixes_string(L"/", dir) || list_contains_string(existing_values, dir)) {
+            if (!string_prefixes_string(L"/", dir) || contains(existing_values, dir)) {
                 any_success = true;
                 continue;
             }
@@ -77,12 +82,10 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
             struct stat buff;
             if (wstat(dir, &buff) == -1) {
                 error = true;
-            }
-            else if (!S_ISDIR(buff.st_mode)) {
+            } else if (!S_ISDIR(buff.st_mode)) {
                 error = true;
                 errno = ENOTDIR;
-            }
-            else if (waccess(dir, X_OK) == -1) {
+            } else if (waccess(dir, X_OK) == -1) {
                 error = true;
             }
 
@@ -90,7 +93,7 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
                 any_success = true;
             } else {
                 streams.err.append_format(_(BUILTIN_SET_PATH_ERROR), L"set", key, dir.c_str(),
-                        strerror(errno));
+                                          strerror(errno));
                 const wchar_t *colon = wcschr(dir.c_str(), L':');
 
                 if (colon && *(colon + 1)) {
@@ -123,6 +126,9 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
     }
 
     switch (env_set(key, val_str, scope | ENV_USER)) {
+        case ENV_OK: {
+            break;
+        }
         case ENV_PERM: {
             streams.err.append_format(_(L"%ls: Tried to change the read-only variable '%ls'\n"),
                                       L"set", key);
@@ -143,6 +149,10 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
             retcode = 1;
             break;
         }
+        default: {
+            DIE("unexpected env_set() ret val");
+            break;
+        }
     }
 
     return retcode;
@@ -159,7 +169,6 @@ static int my_env_set(const wchar_t *key, const wcstring_list_t &val, int scope,
 static int parse_index(std::vector<long> &indexes, const wchar_t *src, const wchar_t *name,
                        size_t var_count, io_streams_t &streams) {
     size_t len;
-
     int count = 0;
     const wchar_t *src_orig = src;
 
@@ -167,9 +176,7 @@ static int parse_index(std::vector<long> &indexes, const wchar_t *src, const wch
         return 0;
     }
 
-    while (*src != L'\0' && (iswalnum(*src) || *src == L'_')) {
-        src++;
-    }
+    while (*src != L'\0' && (iswalnum(*src) || *src == L'_')) src++;
 
     if (*src != L'[') {
         streams.err.append_format(_(BUILTIN_SET_ARG_COUNT), L"set");
@@ -177,7 +184,6 @@ static int parse_index(std::vector<long> &indexes, const wchar_t *src, const wch
     }
 
     len = src - src_orig;
-
     if ((wcsncmp(src_orig, name, len) != 0) || (wcslen(name) != (len))) {
         streams.err.append_format(
             _(L"%ls: Multiple variable names specified in single call (%ls and %.*ls)\n"), L"set",
@@ -186,37 +192,26 @@ static int parse_index(std::vector<long> &indexes, const wchar_t *src, const wch
     }
 
     src++;
-
-    while (iswspace(*src)) {
-        src++;
-    }
+    while (iswspace(*src)) src++;
 
     while (*src != L']') {
-        wchar_t *end;
-
-        long l_ind;
-
-        errno = 0;
-
-        l_ind = wcstol(src, &end, 10);
-
-        if (end == src || errno) {
+        const wchar_t *end;
+        long l_ind = fish_wcstol(src, &end);
+        if (errno > 0) {  // ignore errno == -1 meaning the int did not end with a '\0'
             streams.err.append_format(_(L"%ls: Invalid index starting at '%ls'\n"), L"set", src);
             return 0;
         }
 
-        if (l_ind < 0) {
-            l_ind = var_count + l_ind + 1;
-        }
+        if (l_ind < 0) l_ind = var_count + l_ind + 1;
 
-        src = end;
+        src = end;  //!OCLINT(parameter reassignment)
         if (*src == L'.' && *(src + 1) == L'.') {
             src += 2;
-            long l_ind2 = wcstol(src, &end, 10);
-            if (end == src || errno) {
+            long l_ind2 = fish_wcstol(src, &end);
+            if (errno > 0) {  // ignore errno == -1 meaning the int did not end with a '\0'
                 return 1;
             }
-            src = end;
+            src = end;  //!OCLINT(parameter reassignment)
 
             if (l_ind2 < 0) {
                 l_ind2 = var_count + l_ind2 + 1;
@@ -231,6 +226,7 @@ static int parse_index(std::vector<long> &indexes, const wchar_t *src, const wch
             indexes.push_back(l_ind);
             count++;
         }
+
         while (iswspace(*src)) src++;
     }
 
@@ -319,6 +315,7 @@ static void print_variables(int include_values, int esc, bool shorten_ok, int sc
 
 /// The set builtin creates, updates, and erases (removes, deletes) variables.
 int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+    wchar_t *cmd = argv[0];
     wgetopter_t w;
     // Variables used for parsing the argument list.
     const struct woption long_options[] = {{L"export", no_argument, 0, 'x'},
@@ -342,17 +339,15 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     int erase = 0, list = 0, unexport = 0;
     int universal = 0, query = 0;
     bool shorten_ok = true;
-    bool preserve_incoming_failure_exit_status = true;
+    bool preserve_failure_exit_status = true;
     const int incoming_exit_status = proc_get_last_status();
 
     // Variables used for performing the actual work.
     wchar_t *dest = 0;
-    int retcode = 0;
+    int retcode = STATUS_CMD_OK;
     int scope;
     int slice = 0;
 
-    const wchar_t *bad_char = NULL;
-
     // Parse options to obtain the requested operation and the modifiers.
     w.woptind = 0;
     while (1) {
@@ -368,12 +363,12 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             }
             case 'e': {
                 erase = 1;
-                preserve_incoming_failure_exit_status = false;
+                preserve_failure_exit_status = false;
                 break;
             }
             case 'n': {
                 list = 1;
-                preserve_incoming_failure_exit_status = false;
+                preserve_failure_exit_status = false;
                 break;
             }
             case 'x': {
@@ -402,16 +397,16 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             }
             case 'q': {
                 query = 1;
-                preserve_incoming_failure_exit_status = false;
+                preserve_failure_exit_status = false;
                 break;
             }
             case 'h': {
-                builtin_print_help(parser, streams, argv[0], streams.out);
-                return 0;
+                builtin_print_help(parser, streams, cmd, streams.out);
+                return STATUS_CMD_OK;
             }
             case '?': {
-                builtin_unknown_option(parser, streams, argv[0], argv[w.woptind - 1]);
-                return 1;
+                builtin_unknown_option(parser, streams, cmd, argv[w.woptind - 1]);
+                return STATUS_INVALID_ARGS;
             }
             default: { break; }
         }
@@ -420,32 +415,30 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     // Ok, all arguments have been parsed, let's validate them. If we are checking the existance of
     // a variable (-q) we can not also specify scope.
     if (query && (erase || list)) {
-        streams.err.append_format(BUILTIN_ERR_COMBO, argv[0]);
-
-        builtin_print_help(parser, streams, argv[0], streams.err);
-        return 1;
+        streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
+        builtin_print_help(parser, streams, cmd, streams.err);
+        return STATUS_INVALID_ARGS;
     }
 
     // We can't both list and erase variables.
     if (erase && list) {
-        streams.err.append_format(BUILTIN_ERR_COMBO, argv[0]);
-
-        builtin_print_help(parser, streams, argv[0], streams.err);
-        return 1;
+        streams.err.append_format(BUILTIN_ERR_COMBO, cmd);
+        builtin_print_help(parser, streams, cmd, streams.err);
+        return STATUS_INVALID_ARGS;
     }
 
     // Variables can only have one scope.
     if (local + global + universal > 1) {
-        streams.err.append_format(BUILTIN_ERR_GLOCAL, argv[0]);
-        builtin_print_help(parser, streams, argv[0], streams.err);
-        return 1;
+        streams.err.append_format(BUILTIN_ERR_GLOCAL, cmd);
+        builtin_print_help(parser, streams, cmd, streams.err);
+        return STATUS_INVALID_ARGS;
     }
 
     // Variables can only have one export status.
     if (exportv && unexport) {
         streams.err.append_format(BUILTIN_ERR_EXPUNEXP, argv[0]);
-        builtin_print_help(parser, streams, argv[0], streams.err);
-        return 1;
+        builtin_print_help(parser, streams, cmd, streams.err);
+        return STATUS_INVALID_ARGS;
     }
 
     // Calculate the scope value for variable assignement.
@@ -460,9 +453,8 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             wchar_t *arg = argv[i];
             int slice = 0;
 
-            if (!(dest = wcsdup(arg))) {
-                DIE_MEM();
-            }
+            dest = wcsdup(arg);
+            assert(dest);
 
             if (wcschr(dest, L'[')) {
                 slice = 1;
@@ -478,7 +470,7 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
                 if (!dest_str.missing()) tokenize_variable_array(dest_str, result);
 
                 if (!parse_index(indexes, arg, dest, result.size(), streams)) {
-                    builtin_print_help(parser, streams, argv[0], streams.err);
+                    builtin_print_help(parser, streams, cmd, streams.err);
                     retcode = 1;
                     break;
                 }
@@ -502,16 +494,15 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     if (list) {
         // Maybe we should issue an error if there are any other arguments?
         print_variables(0, 0, shorten_ok, scope, streams);
-        return 0;
+        return STATUS_CMD_OK;
     }
 
     if (w.woptind == argc) {
         // Print values of variables.
         if (erase) {
-            streams.err.append_format(_(L"%ls: Erase needs a variable name\n"), argv[0]);
-
-            builtin_print_help(parser, streams, argv[0], streams.err);
-            retcode = 1;
+            streams.err.append_format(_(L"%ls: Erase needs a variable name\n"), cmd);
+            builtin_print_help(parser, streams, cmd, streams.err);
+            retcode = STATUS_INVALID_ARGS;
         } else {
             print_variables(1, 1, shorten_ok, scope, streams);
         }
@@ -519,27 +510,18 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         return retcode;
     }
 
-    if (!(dest = wcsdup(argv[w.woptind]))) {
-        DIE_MEM();
-    }
+    dest = wcsdup(argv[w.woptind]);
+    assert(dest);
 
     if (wcschr(dest, L'[')) {
         slice = 1;
         *wcschr(dest, L'[') = 0;
     }
 
-    if (!wcslen(dest)) {
-        free(dest);
-        streams.err.append_format(BUILTIN_ERR_VARNAME_ZERO, argv[0]);
+    if (!valid_var_name(dest)) {
+        streams.err.append_format(BUILTIN_ERR_VARNAME, cmd, dest);
         builtin_print_help(parser, streams, argv[0], streams.err);
-        return 1;
-    }
-
-    if ((bad_char = wcsvarname(dest))) {
-        streams.err.append_format(BUILTIN_ERR_VARCHAR, argv[0], *bad_char);
-        builtin_print_help(parser, streams, argv[0], streams.err);
-        free(dest);
-        return 1;
+        return STATUS_INVALID_ARGS;
     }
 
     // Set assignment can work in two modes, either using slices or using the whole array. We detect
@@ -633,7 +615,6 @@ int builtin_set(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
 
     free(dest);
 
-    if (retcode == STATUS_BUILTIN_OK && preserve_incoming_failure_exit_status)
-        retcode = incoming_exit_status;
+    if (retcode == STATUS_CMD_OK && preserve_failure_exit_status) retcode = incoming_exit_status;
     return retcode;
 }
