@@ -1,6 +1,9 @@
 // Functions used for implementing the set_color builtin.
 #include "config.h"
 
+#include <stddef.h>
+#include <stdlib.h>
+
 #if HAVE_NCURSES_H
 #include <ncurses.h>
 #elif HAVE_NCURSES_CURSES_H
@@ -13,9 +16,7 @@
 #elif HAVE_NCURSES_TERM_H
 #include <ncurses/term.h>
 #endif
-#include <assert.h>
-#include <stdlib.h>
-#include <unistd.h>
+
 #include <memory>
 #include <string>
 #include <vector>
@@ -23,9 +24,9 @@
 #include "builtin.h"
 #include "color.h"
 #include "common.h"
+#include "env.h"
 #include "io.h"
 #include "output.h"
-#include "proc.h"
 #include "wgetopt.h"
 #include "wutil.h"  // IWYU pragma: keep
 
@@ -50,18 +51,23 @@ static int set_color_builtin_outputter(char c) {
 
 /// set_color builtin.
 int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
+    // By the time this is called we should have initialized the curses subsystem.
+    assert(curses_initialized);
+
     wgetopter_t w;
     // Variables used for parsing the argument list.
     const struct woption long_options[] = {{L"background", required_argument, 0, 'b'},
                                            {L"help", no_argument, 0, 'h'},
                                            {L"bold", no_argument, 0, 'o'},
                                            {L"underline", no_argument, 0, 'u'},
+                                           {L"italics", no_argument, 0, 'i'},
+                                           {L"dim", no_argument, 0, 'd'},
+                                           {L"reverse", no_argument, 0, 'r'},
                                            {L"version", no_argument, 0, 'v'},
                                            {L"print-colors", no_argument, 0, 'c'},
                                            {0, 0, 0, 0}};
 
-    const wchar_t *short_options = L"b:hvocu";
-
+    const wchar_t *short_options = L"b:hvoidrcu";
     int argc = builtin_count_args(argv);
 
     // Some code passes variables to set_color that don't exist, like $fish_user_whatever. As a
@@ -71,19 +77,18 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     }
 
     const wchar_t *bgcolor = NULL;
-    bool bold = false, underline = false;
-    int errret;
+    bool bold = false, underline = false, italics = false, dim = false, reverse = false;
 
     // Parse options to obtain the requested operation and the modifiers.
     w.woptind = 0;
     while (1) {
-        int c = w.wgetopt_long(argc, argv, short_options, long_options, 0);
+        int opt = w.wgetopt_long(argc, argv, short_options, long_options, 0);
 
-        if (c == -1) {
+        if (opt == -1) {
             break;
         }
 
-        switch (c) {
+        switch (opt) {
             case 0: {
                 break;
             }
@@ -93,22 +98,38 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
             }
             case 'h': {
                 builtin_print_help(parser, streams, argv[0], streams.out);
-                return STATUS_BUILTIN_OK;
+                return STATUS_CMD_OK;
             }
             case 'o': {
                 bold = true;
                 break;
             }
+            case 'i': {
+                italics = true;
+                break;
+            }
+            case 'd': {
+                dim = true;
+                break;
+            }
+            case 'r': {
+                reverse = true;
+                break;
+            }
             case 'u': {
                 underline = true;
                 break;
             }
             case 'c': {
                 print_colors(streams);
-                return STATUS_BUILTIN_OK;
+                return STATUS_CMD_OK;
             }
             case '?': {
-                return STATUS_BUILTIN_ERROR;
+                return STATUS_INVALID_ARGS;
+            }
+            default: {
+                DIE("unexpected opt");
+                break;
             }
         }
     }
@@ -119,14 +140,15 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         rgb_color_t fg = rgb_color_t(argv[w.woptind]);
         if (fg.is_none()) {
             streams.err.append_format(_(L"%ls: Unknown color '%ls'\n"), argv[0], argv[w.woptind]);
-            return STATUS_BUILTIN_ERROR;
+            return STATUS_INVALID_ARGS;
         }
         fgcolors.push_back(fg);
     }
 
-    if (fgcolors.empty() && bgcolor == NULL && !bold && !underline) {
+    if (fgcolors.empty() && bgcolor == NULL && !bold && !underline && !italics && !dim &&
+        !reverse) {
         streams.err.append_format(_(L"%ls: Expected an argument\n"), argv[0]);
-        return STATUS_BUILTIN_ERROR;
+        return STATUS_INVALID_ARGS;
     }
 
     // #1323: We may have multiple foreground colors. Choose the best one. If we had no foreground
@@ -137,19 +159,13 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     const rgb_color_t bg = rgb_color_t(bgcolor ? bgcolor : L"");
     if (bgcolor && bg.is_none()) {
         streams.err.append_format(_(L"%ls: Unknown color '%ls'\n"), argv[0], bgcolor);
-        return STATUS_BUILTIN_ERROR;
-    }
-
-    // Make sure that the term exists.
-    if (cur_term == NULL && setupterm(0, STDOUT_FILENO, &errret) == ERR) {
-        streams.err.append_format(_(L"%ls: Could not set up terminal\n"), argv[0]);
-        return STATUS_BUILTIN_ERROR;
+        return STATUS_INVALID_ARGS;
     }
 
     // Test if we have at least basic support for setting fonts, colors and related bits - otherwise
     // just give up...
     if (cur_term == NULL || !exit_attribute_mode) {
-        return STATUS_BUILTIN_ERROR;
+        return STATUS_CMD_ERROR;
     }
 
     // Save old output function so we can restore it.
@@ -159,19 +175,31 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     builtin_set_color_output.clear();
     output_set_writer(set_color_builtin_outputter);
 
-    if (bold) {
-        if (enter_bold_mode) writembs(tparm(enter_bold_mode));
+    if (bold && enter_bold_mode) {
+        writembs(tparm(enter_bold_mode));
     }
 
-    if (underline) {
-        if (enter_underline_mode) writembs(enter_underline_mode);
+    if (underline && enter_underline_mode) {
+        writembs(enter_underline_mode);
     }
 
-    if (bgcolor != NULL) {
-        if (bg.is_normal()) {
-            write_color(rgb_color_t::black(), false /* not is_fg */);
-            writembs(tparm(exit_attribute_mode));
-        }
+    if (italics && enter_italics_mode) {
+        writembs(enter_italics_mode);
+    }
+
+    if (dim && enter_dim_mode) {
+        writembs(enter_dim_mode);
+    }
+
+    if (reverse && enter_reverse_mode) {
+        writembs(enter_reverse_mode);
+    } else if (reverse && enter_standout_mode) {
+        writembs(enter_standout_mode);
+    }
+
+    if (bgcolor != NULL && bg.is_normal()) {
+        write_color(rgb_color_t::black(), false /* not is_fg */);
+        writembs(tparm(exit_attribute_mode));
     }
 
     if (!fg.is_none()) {
@@ -188,10 +216,8 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
         }
     }
 
-    if (bgcolor != NULL) {
-        if (!bg.is_normal() && !bg.is_reset()) {
-            write_color(bg, false /* not is_fg */);
-        }
+    if (bgcolor != NULL && !bg.is_normal() && !bg.is_reset()) {
+        write_color(bg, false /* not is_fg */);
     }
 
     // Restore saved writer function.
@@ -201,5 +227,5 @@ int builtin_set_color(parser_t &parser, io_streams_t &streams, wchar_t **argv) {
     streams.out.append(str2wcstring(builtin_set_color_output));
     builtin_set_color_output.clear();
 
-    return STATUS_BUILTIN_OK;
+    return STATUS_CMD_OK;
 }
