@@ -1,12 +1,11 @@
 // Various functions, mostly string utilities, that are used by most parts of fish.
 #include "config.h"
 
-#include <assert.h>
 #include <cxxabi.h>
 #include <dlfcn.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <limits.h>
-#include <signal.h>
 #include <stdarg.h>
 #include <stddef.h>
 #include <stdint.h>
@@ -28,25 +27,30 @@
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
+
 #include <algorithm>
 #include <memory>  // IWYU pragma: keep
+#include <type_traits>
 
 #include "common.h"
+#include "env.h"
 #include "expand.h"
 #include "fallback.h"  // IWYU pragma: keep
+#include "proc.h"
 #include "wildcard.h"
 #include "wutil.h"  // IWYU pragma: keep
 
-#define NOT_A_WCHAR (static_cast<wint_t>(WEOF))
+constexpr wint_t NOT_A_WCHAR = static_cast<wint_t>(WEOF);
 
 struct termios shell_modes;
 
 // Note we foolishly assume that pthread_t is just a primitive. But it might be a struct.
 static pthread_t main_thread_id = 0;
-static bool thread_assertions_configured_for_testing = false;
+static bool thread_asserts_cfg_for_testing = false;
 
 wchar_t ellipsis_char;
 wchar_t omitted_newline_char;
+wchar_t obfuscation_read_char;
 bool g_profiling_active = false;
 const wchar_t *program_name;
 int debug_level = 1;         // default maximum debug output level (errors and warnings)
@@ -56,14 +60,14 @@ int debug_stack_frames = 0;  // default number of stack frames to show on debug(
 static pid_t initial_pid = 0;
 
 /// Be able to restore the term's foreground process group.
-static pid_t initial_foreground_process_group = -1;
+static pid_t initial_fg_process_group = -1;
 
 /// This struct maintains the current state of the terminal size. It is updated on demand after
 /// receiving a SIGWINCH. Do not touch this struct directly, it's managed with a rwlock. Use
 /// common_get_width()/common_get_height().
-static struct winsize termsize;
-static volatile bool termsize_valid;
-static rwlock_t termsize_rwlock;
+static pthread_mutex_t termsize_lock = PTHREAD_MUTEX_INITIALIZER;
+static struct winsize termsize = {USHRT_MAX, USHRT_MAX, USHRT_MAX, USHRT_MAX};
+static volatile bool termsize_valid = false;
 
 static char *wcs2str_internal(const wchar_t *in, char *out);
 static void debug_shared(const wchar_t msg_level, const wcstring &msg);
@@ -106,7 +110,7 @@ demangled_backtrace(int max_frames, int skip_levels) {
 
 void __attribute__((noinline))
 show_stackframe(const wchar_t msg_level, int frame_count, int skip_levels) {
-    ASSERT_IS_NOT_FORKED_CHILD();
+    if (frame_count < 1) return;
 
     // TODO: Decide if this is still needed. I'm commenting it out because it caused me some grief
     // while trying to debug a test failure. And the tests run just fine without spurious failures
@@ -115,7 +119,6 @@ show_stackframe(const wchar_t msg_level, int frame_count, int skip_levels) {
     // Hack to avoid showing backtraces in the tester.
     // if (program_name && !wcscmp(program_name, L"(ignore)")) return;
 
-    if (frame_count < 1) frame_count = 999;
     debug_shared(msg_level, L"Backtrace:");
     std::vector<wcstring> bt = demangled_backtrace(frame_count, skip_levels + 2);
     for (int i = 0; (size_t)i < bt.size(); i++) {
@@ -196,15 +199,14 @@ static wcstring str2wcs_internal(const char *in, const size_t in_len) {
             // Protect against broken mbrtowc() implementations which attempt to encode UTF-8
             // sequences longer than four bytes (e.g., OS X Snow Leopard).
             use_encode_direct = true;
-        } else if (sizeof(wchar_t) == 2 && (in[in_pos] & 0xF8) == 0xF0) {
+        } else if (sizeof(wchar_t) == 2 &&  //!OCLINT(constant if expression)
+                   (in[in_pos] & 0xF8) == 0xF0) {
             // Assume we are in a UTF-16 environment (e.g., Cygwin) using a UTF-8 encoding.
             // The bits set check will be true for a four byte UTF-8 sequence that requires
             // two UTF-16 chars. Something that doesn't work with our simple use of mbrtowc().
             use_encode_direct = true;
         } else {
             ret = mbrtowc(&wc, &in[in_pos], in_len - in_pos, &state);
-            // fprintf(stderr, "WTF in_pos %d  ret %d\n", in_pos, ret);
-
             // Determine whether to encode this character with our crazy scheme.
             if (wc >= ENCODE_DIRECT_BASE && wc < ENCODE_DIRECT_BASE + 256) {
                 use_encode_direct = true;
@@ -219,7 +221,8 @@ static wcstring str2wcs_internal(const char *in, const size_t in_len) {
             } else if (ret > in_len - in_pos) {
                 // Other error codes? Terrifying, should never happen.
                 use_encode_direct = true;
-            } else if (sizeof(wchar_t) == 2 && wc >= 0xD800 && wc <= 0xDFFF) {
+            } else if (sizeof(wchar_t) == 2 && wc >= 0xD800 &&  //!OCLINT(constant if expression)
+                       wc <= 0xDFFF) {
                 // If we get a surrogate pair char on a UTF-16 system (e.g., Cygwin) then
                 // it's guaranteed the UTF-8 decoding is wrong so use direct encoding.
                 use_encode_direct = true;
@@ -227,24 +230,20 @@ static wcstring str2wcs_internal(const char *in, const size_t in_len) {
         }
 
         if (use_encode_direct) {
-            // fprintf(stderr, "WTF use_encode_direct\n");
             wc = ENCODE_DIRECT_BASE + (unsigned char)in[in_pos];
             result.push_back(wc);
             in_pos++;
             memset(&state, 0, sizeof state);
-        } else if (ret == 0) {
-            // fprintf(stderr, "WTF null byte\n");
-            // Embedded null byte!
+        } else if (ret == 0) {  // embedded null byte!
             result.push_back(L'\0');
             in_pos++;
             memset(&state, 0, sizeof state);
-        } else {
-            // fprintf(stderr, "WTF null byte\n");
-            // Normal case.
+        } else {  // normal case
             result.push_back(wc);
             in_pos += ret;
         }
     }
+
     return result;
 }
 
@@ -267,14 +266,14 @@ char *wcs2str(const wchar_t *in) {
         if (result) {
             // It converted into the local buffer, so copy it.
             result = strdup(result);
-            if (!result) DIE_MEM();
+            assert(result);
         }
         return result;
     }
 
     // Here we probably allocate a buffer probably much larger than necessary.
     char *out = (char *)malloc(MAX_UTF8_BYTES * wcslen(in) + 1);
-    if (!out) DIE_MEM();
+    assert(out);
     return wcs2str_internal(in, out);
 }
 
@@ -286,16 +285,15 @@ std::string wcs2string(const wcstring &input) {
     result.reserve(input.size());
 
     mbstate_t state = {};
-    char converted[MB_LEN_MAX + 1];
+    char converted[MB_LEN_MAX];
 
     for (size_t i = 0; i < input.size(); i++) {
         wchar_t wc = input[i];
         if (wc == INTERNAL_SEPARATOR) {
-            // Do nothing.
+            ;  // do nothing
         } else if (wc >= ENCODE_DIRECT_BASE && wc < ENCODE_DIRECT_BASE + 256) {
             result.push_back(wc - ENCODE_DIRECT_BASE);
-        } else if (MB_CUR_MAX == 1)  // single-byte locale (C/POSIX/ISO-8859)
-        {
+        } else if (MB_CUR_MAX == 1) {  // single-byte locale (C/POSIX/ISO-8859)
             // If `wc` contains a wide character we emit a question-mark.
             if (wc & ~0xFF) {
                 wc = '?';
@@ -305,8 +303,8 @@ std::string wcs2string(const wcstring &input) {
         } else {
             memset(converted, 0, sizeof converted);
             size_t len = wcrtomb(converted, wc, &state);
-            if (len == (size_t)(-1)) {
-                debug(1, L"Wide character %d has no narrow representation", wc);
+            if (len == (size_t)-1) {
+                debug(1, L"Wide character U+%4X has no narrow representation", wc);
                 memset(&state, 0, sizeof(state));
             } else {
                 result.append(converted, len);
@@ -332,7 +330,7 @@ static char *wcs2str_internal(const wchar_t *in, char *out) {
 
     while (in[in_pos]) {
         if (in[in_pos] == INTERNAL_SEPARATOR) {
-            // Do nothing.
+            ;  // do nothing
         } else if (in[in_pos] >= ENCODE_DIRECT_BASE && in[in_pos] < ENCODE_DIRECT_BASE + 256) {
             out[out_pos++] = in[in_pos] - ENCODE_DIRECT_BASE;
         } else if (MB_CUR_MAX == 1)  // single-byte locale (C/POSIX/ISO-8859)
@@ -346,7 +344,7 @@ static char *wcs2str_internal(const wchar_t *in, char *out) {
         } else {
             size_t len = wcrtomb(&out[out_pos], in[in_pos], &state);
             if (len == (size_t)-1) {
-                debug(1, L"Wide character %d has no narrow representation", in[in_pos]);
+                debug(1, L"Wide character U+%4X has no narrow representation", in[in_pos]);
                 memset(&state, 0, sizeof(state));
             } else {
                 out_pos += len;
@@ -359,6 +357,14 @@ static char *wcs2str_internal(const wchar_t *in, char *out) {
     return out;
 }
 
+/// Test if the character can be encoded using the current locale.
+static bool can_be_encoded(wchar_t wc) {
+    char converted[MB_LEN_MAX];
+    mbstate_t state = {};
+
+    return wcrtomb(converted, wc, &state) != (size_t)-1;
+}
+
 wcstring format_string(const wchar_t *format, ...) {
     va_list va;
     va_start(va, format);
@@ -394,9 +400,7 @@ void append_formatv(wcstring &target, const wchar_t *format, va_list va_orig) {
                 break;
             }
             buff = (wchar_t *)realloc((buff == static_buff ? NULL : buff), size);
-            if (buff == NULL) {
-                DIE_MEM();
-            }
+            assert(buff != NULL);
         }
 
         // Try printing.
@@ -449,64 +453,38 @@ wchar_t *quote_end(const wchar_t *pos) {
 }
 
 void fish_setlocale() {
-    // Use ellipsis if on known unicode system, otherwise use $.
-    ellipsis_char = (fish_wcwidth(L'\x2026') > 0) ? L'\x2026' : L'$';
+    // Use the Unicode "ellipsis" symbol if it can be encoded using the current locale.
+    ellipsis_char = can_be_encoded(L'\x2026') ? L'\x2026' : L'$';
 
-    // U+23CE is the "return" character
-    omitted_newline_char = (fish_wcwidth(L'\x23CE') > 0) ? L'\x23CE' : L'~';
-}
+    // Use the Unicode "return" symbol if it can be encoded using the current locale.
+    omitted_newline_char = can_be_encoded(L'\x23CE') ? L'\x23CE' : L'~';
 
-bool contains_internal(const wchar_t *a, int vararg_handle, ...) {
-    const wchar_t *arg;
-    va_list va;
-    bool res = false;
-
-    CHECK(a, 0);
-
-    va_start(va, vararg_handle);
-    while ((arg = va_arg(va, const wchar_t *)) != 0) {
-        if (wcscmp(a, arg) == 0) {
-            res = true;
-            break;
-        }
-    }
-    va_end(va);
-    return res;
-}
-
-/// wcstring variant of contains_internal. The first parameter is a wcstring, the rest are const
-/// wchar_t *. vararg_handle exists only to give us a POD-value to pass to va_start.
-__sentinel bool contains_internal(const wcstring &needle, int vararg_handle, ...) {
-    const wchar_t *arg;
-    va_list va;
-    int res = 0;
-
-    const wchar_t *needle_cstr = needle.c_str();
-    va_start(va, vararg_handle);
-    while ((arg = va_arg(va, const wchar_t *)) != 0) {
-        // libc++ has an unfortunate implementation of operator== that unconditonally wcslen's the
-        // wchar_t* parameter, so prefer wcscmp directly.
-        if (!wcscmp(needle_cstr, arg)) {
-            res = 1;
-            break;
-        }
-    }
-    va_end(va);
-    return res;
+    // solid circle unicode character if it is able, fallback to the hash character
+    obfuscation_read_char = can_be_encoded(L'\u25cf') ? L'\u25cf' : L'#';
 }
 
 long read_blocked(int fd, void *buf, size_t count) {
-    ssize_t res;
-    sigset_t chldset, oldset;
+    long bytes_read = 0;
 
-    sigemptyset(&chldset);
-    sigaddset(&chldset, SIGCHLD);
-    VOMIT_ON_FAILURE(pthread_sigmask(SIG_BLOCK, &chldset, &oldset));
-    res = read(fd, buf, count);
-    VOMIT_ON_FAILURE(pthread_sigmask(SIG_SETMASK, &oldset, NULL));
-    return res;
+    while (count) {
+        ssize_t res = read(fd, (char *)buf + bytes_read, count);
+        if (res == 0) {
+            break;
+        } else if (res == -1) {
+            if (errno == EINTR) continue;
+            if (errno == EAGAIN) return bytes_read ? bytes_read : -1;
+            return -1;
+        } else {
+            bytes_read += res;
+            count -= res;
+        }
+    }
+
+    return bytes_read;
 }
 
+/// Loop a write request while failure is non-critical. Return -1 and set errno in case of critical
+/// error.
 ssize_t write_loop(int fd, const char *buff, size_t count) {
     size_t out_cum = 0;
     while (out_cum < count) {
@@ -530,15 +508,20 @@ ssize_t read_loop(int fd, void *buff, size_t count) {
     return result;
 }
 
-static bool should_debug(int level) {
-    if (level > debug_level) return false;
-
-    // Hack to not print error messages in the tests.
-    if (program_name && !wcscmp(program_name, L"(ignore)")) return false;
-
-    return true;
+/// Hack to not print error messages in the tests. Do not call this from functions in this module
+/// like `debug()`. It is only intended to supress diagnostic noise from testing things like the
+/// fish parser where we expect a lot of diagnostic messages due to testing error conditions.
+bool should_suppress_stderr_for_tests() {
+    return program_name && !wcscmp(program_name, TESTS_PROGRAM_NAME);
 }
 
+/// Return true if we should emit a `debug()` message. This used to call
+/// `should_suppress_stderr_for_tests()`. It no longer does so because it can suppress legitimate
+/// errors we want to see if things go wrong. Too, calling that function is no longer necessary, if
+/// it ever was, to suppress unwanted diagnostic output that might confuse people running `make
+/// test`.
+static bool should_debug(int level) { return level <= debug_level; }
+
 static void debug_shared(const wchar_t level, const wcstring &msg) {
     pid_t current_pid = getpid();
 
@@ -600,7 +583,7 @@ void debug_safe(int level, const char *msg, const char *param1, const char *para
                                   param7, param8, param9, param10, param11, param12};
     if (!msg) return;
 
-    // Can't call printf, that may allocate memory Just call write() over and over.
+    // Can't call fwprintf, that may allocate memory Just call write() over and over.
     if (level > debug_level) return;
     int errno_old = errno;
 
@@ -768,9 +751,9 @@ static void escape_string_internal(const wchar_t *orig_in, size_t in_len, wcstri
     assert(orig_in != NULL);
 
     const wchar_t *in = orig_in;
-    bool escape_all = !!(flags & ESCAPE_ALL);
-    bool no_quoted = !!(flags & ESCAPE_NO_QUOTED);
-    bool no_tilde = !!(flags & ESCAPE_NO_TILDE);
+    bool escape_all = static_cast<bool>(flags & ESCAPE_ALL);
+    bool no_quoted = static_cast<bool>(flags & ESCAPE_NO_QUOTED);
+    bool no_tilde = static_cast<bool>(flags & ESCAPE_NO_TILDE);
 
     int need_escape = 0;
     int need_complex_escape = 0;
@@ -825,7 +808,7 @@ static void escape_string_internal(const wchar_t *orig_in, size_t in_len, wcstri
                     need_escape = need_complex_escape = 1;
                     break;
                 }
-                case L'\x1b': {
+                case L'\e': {
                     out += L'\\';
                     out += L'e';
                     need_escape = need_complex_escape = 1;
@@ -834,7 +817,8 @@ static void escape_string_internal(const wchar_t *orig_in, size_t in_len, wcstri
                 case L'\\':
                 case L'\'': {
                     need_escape = need_complex_escape = 1;
-                    if (escape_all) out += L'\\';
+                    // WTF if (escape_all) out += L'\\';
+                    out += L'\\';
                     out += *in;
                     break;
                 }
@@ -919,7 +903,7 @@ static void escape_string_internal(const wchar_t *orig_in, size_t in_len, wcstri
     }
 }
 
-wcstring escape(const wchar_t *in, escape_flags_t flags) {
+wcstring escape_string(const wchar_t *in, escape_flags_t flags) {
     wcstring result;
     escape_string_internal(in, wcslen(in), &result, flags);
     return result;
@@ -933,8 +917,7 @@ wcstring escape_string(const wcstring &in, escape_flags_t flags) {
 
 /// Helper to return the last character in a string, or NOT_A_WCHAR.
 static wint_t string_last_char(const wcstring &str) {
-    size_t len = str.size();
-    return len == 0 ? NOT_A_WCHAR : str.at(len - 1);
+    return str.empty() ? NOT_A_WCHAR : str.back();
 }
 
 /// Given a null terminated string starting with a backslash, read the escape as if it is unquoted,
@@ -1058,9 +1041,9 @@ size_t read_unquoted_escape(const wchar_t *input, wcstring *result, bool allow_i
             }
             break;
         }
-        // \x1b means escape.
+        // \e means escape.
         case L'e': {
-            result_char_or_none = L'\x1b';
+            result_char_or_none = L'\e';
             break;
         }
         // \f means form feed.
@@ -1117,8 +1100,8 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
     wcstring result;
     result.reserve(input_len);
 
-    const bool unescape_special = !!(flags & UNESCAPE_SPECIAL);
-    const bool allow_incomplete = !!(flags & UNESCAPE_INCOMPLETE);
+    const bool unescape_special = static_cast<bool>(flags & UNESCAPE_SPECIAL);
+    const bool allow_incomplete = static_cast<bool>(flags & UNESCAPE_INCOMPLETE);
 
     int bracket_count = 0;
 
@@ -1212,14 +1195,15 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                 }
                 case L'\'': {
                     mode = mode_single_quotes;
-                    to_append_or_none = unescape_special ? INTERNAL_SEPARATOR : NOT_A_WCHAR;
+                    to_append_or_none = unescape_special ? wint_t(INTERNAL_SEPARATOR) : NOT_A_WCHAR;
                     break;
                 }
                 case L'\"': {
                     mode = mode_double_quotes;
-                    to_append_or_none = unescape_special ? INTERNAL_SEPARATOR : NOT_A_WCHAR;
+                    to_append_or_none = unescape_special ? wint_t(INTERNAL_SEPARATOR) : NOT_A_WCHAR;
                     break;
                 }
+                default: { break; }
             }
         } else if (mode == mode_single_quotes) {
             if (c == L'\\') {
@@ -1250,14 +1234,14 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                     }
                 }
             } else if (c == L'\'') {
-                to_append_or_none = unescape_special ? INTERNAL_SEPARATOR : NOT_A_WCHAR;
+                to_append_or_none = unescape_special ? wint_t(INTERNAL_SEPARATOR) : NOT_A_WCHAR;
                 mode = mode_unquoted;
             }
         } else if (mode == mode_double_quotes) {
             switch (c) {
                 case L'"': {
                     mode = mode_unquoted;
-                    to_append_or_none = unescape_special ? INTERNAL_SEPARATOR : NOT_A_WCHAR;
+                    to_append_or_none = unescape_special ? wint_t(INTERNAL_SEPARATOR) : NOT_A_WCHAR;
                     break;
                 }
                 case '\\': {
@@ -1297,6 +1281,7 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
                     }
                     break;
                 }
+                default: { break; }
             }
         }
 
@@ -1311,7 +1296,7 @@ static bool unescape_string_internal(const wchar_t *const input, const size_t in
 
     // Return the string by reference, and then success.
     if (!errored) {
-        output_str->swap(result);
+        *output_str = std::move(result);
     }
     return !errored;
 }
@@ -1321,7 +1306,7 @@ bool unescape_string_in_place(wcstring *str, unescape_flags_t escape_special) {
     wcstring output;
     bool success = unescape_string_internal(str->c_str(), str->size(), &output, escape_special);
     if (success) {
-        str->swap(output);
+        *str = std::move(output);
     }
     return success;
 }
@@ -1338,32 +1323,100 @@ bool unescape_string(const wcstring &input, wcstring *output, unescape_flags_t e
     return success;
 }
 
-void common_handle_winch(int signal) {
-    // Don't run ioctl() here, it's not safe to use in signals.
-    UNUSED(signal);
+/// Used to invalidate our idea of having a valid window size. This can occur when either the
+/// COLUMNS or LINES variables are changed. This is also invoked when the shell regains control of
+/// the tty since it is possible the terminal size changed while an external command was running.
+void invalidate_termsize(bool invalidate_vars) {
     termsize_valid = false;
+    if (invalidate_vars) {
+        termsize.ws_col = termsize.ws_row = USHRT_MAX;
+    }
+}
+
+/// Handle SIGWINCH. This is also invoked when the shell regains control of the tty since it is
+/// possible the terminal size changed while an external command was running.
+void common_handle_winch(int signal) {
+    // Don't run ioctl() here. Technically it's not safe to use in signals although in practice it
+    // is safe on every platform I've used. But we want to be conservative on such matters.
+    UNUSED(signal);
+    invalidate_termsize(false);
+}
+
+/// Validate the new terminal size. Fallback to the env vars if necessary. Ensure the values are
+/// sane and if not fallback to a default of 80x24.
+static void validate_new_termsize(struct winsize *new_termsize) {
+    if (new_termsize->ws_col == 0 || new_termsize->ws_row == 0) {
+#ifdef HAVE_WINSIZE
+        if (shell_is_interactive()) {
+            debug(1, _(L"Current terminal parameters have rows and/or columns set to zero."));
+            debug(1, _(L"The stty command can be used to correct this "
+                       L"(e.g., stty rows 80 columns 24)."));
+        }
+#endif
+        // Fallback to the environment vars.
+        env_var_t col_var = env_get_string(L"COLUMNS");
+        env_var_t row_var = env_get_string(L"LINES");
+        if (!col_var.missing_or_empty() && !row_var.missing_or_empty()) {
+            // Both vars have to have valid values.
+            int col = fish_wcstoi(col_var.c_str());
+            bool col_ok = errno == 0 && col > 0 && col <= USHRT_MAX;
+            int row = fish_wcstoi(row_var.c_str());
+            bool row_ok = errno == 0 && row > 0 && row <= USHRT_MAX;
+            if (col_ok && row_ok) {
+                new_termsize->ws_col = col;
+                new_termsize->ws_row = row;
+            }
+        }
+    }
+
+    if (new_termsize->ws_col < MIN_TERM_COL || new_termsize->ws_row < MIN_TERM_ROW) {
+        if (shell_is_interactive()) {
+            debug(1, _(L"Current terminal parameters set terminal size to unreasonable value."));
+            debug(1, _(L"Defaulting terminal size to 80x24."));
+        }
+        new_termsize->ws_col = DFLT_TERM_COL;
+        new_termsize->ws_row = DFLT_TERM_ROW;
+    }
+}
+
+/// Export the new terminal size as env vars and to the kernel if possible.
+static void export_new_termsize(struct winsize *new_termsize) {
+    wchar_t buf[64];
+    env_var_t cols = env_get_string(L"COLUMNS", ENV_EXPORT);
+    swprintf(buf, 64, L"%d", (int)new_termsize->ws_col);
+    env_set(L"COLUMNS", buf, ENV_GLOBAL | (cols.missing_or_empty() ? 0 : ENV_EXPORT));
+
+    env_var_t lines = env_get_string(L"LINES", ENV_EXPORT);
+    swprintf(buf, 64, L"%d", (int)new_termsize->ws_row);
+    env_set(L"LINES", buf, ENV_GLOBAL | (lines.missing_or_empty() ? 0 : ENV_EXPORT));
+
+#ifdef HAVE_WINSIZE
+    ioctl(STDOUT_FILENO, TIOCSWINSZ, new_termsize);
+#endif
 }
 
 /// Updates termsize as needed, and returns a copy of the winsize.
-static struct winsize get_current_winsize() {
-#ifndef HAVE_WINSIZE
-    struct winsize retval = {0};
-    retval.ws_col = 80;
-    retval.ws_row = 24;
-    return retval;
-#endif
-    scoped_rwlock guard(termsize_rwlock, true);
-    struct winsize retval = termsize;
-    if (!termsize_valid) {
-        struct winsize size;
-        if (ioctl(1, TIOCGWINSZ, &size) == 0) {
-            retval = size;
-            guard.upgrade();
-            termsize = retval;
-        }
+struct winsize get_current_winsize() {
+    scoped_lock guard(termsize_lock);
+
+    if (termsize_valid) return termsize;
+
+    struct winsize new_termsize = {0, 0, 0, 0};
+#ifdef HAVE_WINSIZE
+    errno = 0;
+    if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &new_termsize) != -1 &&
+        new_termsize.ws_col == termsize.ws_col && new_termsize.ws_row == termsize.ws_row) {
         termsize_valid = true;
+        return termsize;
     }
-    return retval;
+#endif
+
+    validate_new_termsize(&new_termsize);
+    export_new_termsize(&new_termsize);
+    termsize.ws_col = new_termsize.ws_col;
+    termsize.ws_row = new_termsize.ws_row;
+    termsize_valid = true;
+    return termsize;
 }
 
 int common_get_width() { return get_current_winsize().ws_col; }
@@ -1500,12 +1553,12 @@ int string_fuzzy_match_t::compare(const string_fuzzy_match_t &rhs) const {
     return 0;  // equal
 }
 
-bool list_contains_string(const wcstring_list_t &list, const wcstring &str) {
+bool contains(const wcstring_list_t &list, const wcstring &str) {
     return std::find(list.begin(), list.end(), str) != list.end();
 }
 
 int create_directory(const wcstring &d) {
-    int ok = 0;
+    bool ok = false;
     struct stat buf;
     int stat_res = 0;
 
@@ -1514,27 +1567,18 @@ int create_directory(const wcstring &d) {
     }
 
     if (stat_res == 0) {
-        if (S_ISDIR(buf.st_mode)) {
-            ok = 1;
-        }
-    } else {
-        if (errno == ENOENT) {
-            wcstring dir = wdirname(d);
-            if (!create_directory(dir)) {
-                if (!wmkdir(d, 0700)) {
-                    ok = 1;
-                }
-            }
-        }
+        if (S_ISDIR(buf.st_mode)) ok = true;
+    } else if (errno == ENOENT) {
+        wcstring dir = wdirname(d);
+        if (!create_directory(dir) && !wmkdir(d, 0700)) ok = true;
     }
 
     return ok ? 0 : -1;
 }
 
 __attribute__((noinline)) void bugreport() {
-    debug(1, _(L"This is a bug. Break on bugreport to debug. "
-               L"If you can reproduce it, please send a bug report to %s."),
-          PACKAGE_BUGREPORT);
+    debug(0, _(L"This is a bug. Break on bugreport to debug."));
+    debug(0, _(L"If you can reproduce it, please send a bug report to %s."), PACKAGE_BUGREPORT);
 }
 
 wcstring format_size(long long sz) {
@@ -1632,7 +1676,7 @@ void format_size_safe(char buff[128], unsigned long long sz) {
 /// the gettimeofday function and will have the same precision as that function.
 double timef() {
     struct timeval tv;
-    VOMIT_ON_FAILURE(gettimeofday(&tv, 0));
+    assert_with_errno(gettimeofday(&tv, 0) != -1);
     // return (double)tv.tv_sec + 0.000001 * tv.tv_usec;
     return (double)tv.tv_sec + 1e-6 * tv.tv_usec;
 }
@@ -1682,9 +1726,7 @@ __attribute__((noinline)) void debug_thread_error(void) {
 
 void set_main_thread() { main_thread_id = pthread_self(); }
 
-void configure_thread_assertions_for_testing(void) {
-    thread_assertions_configured_for_testing = true;
-}
+void configure_thread_assertions_for_testing(void) { thread_asserts_cfg_for_testing = true; }
 
 bool is_forked_child(void) {
     // Just bail if nobody's called setup_fork_guards, e.g. some of our tools.
@@ -1692,7 +1734,7 @@ bool is_forked_child(void) {
 
     bool is_child_of_fork = (getpid() != initial_pid);
     if (is_child_of_fork) {
-        printf("Uh-oh: %d\n", getpid());
+        debug(0, L"Uh-oh: getpid() != initial_pid: %d != %d\n", getpid(), initial_pid);
         while (1) sleep(10000);
     }
     return is_child_of_fork;
@@ -1704,14 +1746,15 @@ void setup_fork_guards(void) {
 }
 
 void save_term_foreground_process_group(void) {
-    initial_foreground_process_group = tcgetpgrp(STDIN_FILENO);
+    initial_fg_process_group = tcgetpgrp(STDIN_FILENO);
 }
 
 void restore_term_foreground_process_group(void) {
-    if (initial_foreground_process_group != -1) {
-        // This is called during shutdown and from a signal handler. We don't bother to complain on
-        // failure.
-        tcsetpgrp(STDIN_FILENO, initial_foreground_process_group);
+    if (initial_fg_process_group == -1) return;
+    // This is called during shutdown and from a signal handler. We don't bother to complain on
+    // failure because doing so is unlikely to be noticed.
+    if (tcsetpgrp(STDIN_FILENO, initial_fg_process_group) == -1 && errno == ENOTTY) {
+        redirect_tty_output();
     }
 }
 
@@ -1721,29 +1764,25 @@ bool is_main_thread() {
 }
 
 void assert_is_main_thread(const char *who) {
-    if (!is_main_thread() && !thread_assertions_configured_for_testing) {
-        fprintf(stderr,
-                "Warning: %s called off of main thread. Break on debug_thread_error to debug.\n",
-                who);
+    if (!is_main_thread() && !thread_asserts_cfg_for_testing) {
+        debug(0, "%s called off of main thread.", who);
+        debug(0, "Break on debug_thread_error to debug.");
         debug_thread_error();
     }
 }
 
 void assert_is_not_forked_child(const char *who) {
     if (is_forked_child()) {
-        fprintf(stderr,
-                "Warning: %s called in a forked child. Break on debug_thread_error to debug.\n",
-                who);
+        debug(0, "%s called in a forked child.", who);
+        debug(0, "Break on debug_thread_error to debug.");
         debug_thread_error();
     }
 }
 
 void assert_is_background_thread(const char *who) {
-    if (is_main_thread() && !thread_assertions_configured_for_testing) {
-        fprintf(stderr,
-                "Warning: %s called on the main thread (may block!). Break on debug_thread_error "
-                "to debug.\n",
-                who);
+    if (is_main_thread() && !thread_asserts_cfg_for_testing) {
+        debug(0, "%s called on the main thread (may block!).", who);
+        debug(0, "Break on debug_thread_error to debug.");
         debug_thread_error();
     }
 }
@@ -1751,26 +1790,24 @@ void assert_is_background_thread(const char *who) {
 void assert_is_locked(void *vmutex, const char *who, const char *caller) {
     pthread_mutex_t *mutex = static_cast<pthread_mutex_t *>(vmutex);
     if (0 == pthread_mutex_trylock(mutex)) {
-        fprintf(stderr,
-                "Warning: %s is not locked when it should be in '%s'. Break on debug_thread_error "
-                "to debug.\n",
-                who, caller);
+        debug(0, "%s is not locked when it should be in '%s'", who, caller);
+        debug(0, "Break on debug_thread_error to debug.");
         debug_thread_error();
         pthread_mutex_unlock(mutex);
     }
 }
 
 void scoped_lock::lock(void) {
-    assert(!locked);
+    assert(!locked);  //!OCLINT(multiple unary operator)
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_mutex_lock(lock_obj));
+    DIE_ON_FAILURE(pthread_mutex_lock(lock_obj));
     locked = true;
 }
 
 void scoped_lock::unlock(void) {
     assert(locked);
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_mutex_unlock(lock_obj));
+    DIE_ON_FAILURE(pthread_mutex_unlock(lock_obj));
     locked = false;
 }
 
@@ -1785,41 +1822,44 @@ scoped_lock::~scoped_lock() {
 }
 
 void scoped_rwlock::lock(void) {
-    assert(!(locked || locked_shared));
+    assert(!(locked || locked_shared));  //!OCLINT(multiple unary operator)
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_rdlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_rdlock(rwlock_obj));
     locked = true;
 }
 
 void scoped_rwlock::unlock(void) {
     assert(locked);
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_unlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_unlock(rwlock_obj));
     locked = false;
 }
 
 void scoped_rwlock::lock_shared(void) {
-    assert(!(locked || locked_shared));
+    assert(!(locked || locked_shared));  //!OCLINT(multiple unary operator)
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_wrlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_wrlock(rwlock_obj));
     locked_shared = true;
 }
 
 void scoped_rwlock::unlock_shared(void) {
     assert(locked_shared);
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_unlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_unlock(rwlock_obj));
     locked_shared = false;
 }
 
+#if 0
+// This is not currently used.
 void scoped_rwlock::upgrade(void) {
     assert(locked_shared);
     ASSERT_IS_NOT_FORKED_CHILD();
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_unlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_unlock(rwlock_obj));
     locked = false;
-    VOMIT_ON_FAILURE_NO_ERRNO(pthread_rwlock_wrlock(rwlock_obj));
+    DIE_ON_FAILURE(pthread_rwlock_wrlock(rwlock_obj));
     locked_shared = true;
 }
+#endif
 
 scoped_rwlock::scoped_rwlock(pthread_rwlock_t &rwlock, bool shared)
     : rwlock_obj(&rwlock), locked(false), locked_shared(false) {
@@ -1847,22 +1887,6 @@ scoped_rwlock::~scoped_rwlock() {
     }
 }
 
-wcstokenizer::wcstokenizer(const wcstring &s, const wcstring &separator)
-    : buffer(), str(), state(), sep(separator) {
-    buffer = wcsdup(s.c_str());
-    str = buffer;
-    state = NULL;
-}
-
-bool wcstokenizer::next(wcstring &result) {
-    wchar_t *tmp = wcstok(str, sep.c_str(), &state);
-    str = NULL;
-    if (tmp) result = tmp;
-    return tmp != NULL;
-}
-
-wcstokenizer::~wcstokenizer() { free(buffer); }
-
 template <typename CharType_t>
 static CharType_t **make_null_terminated_array_helper(
     const std::vector<std::basic_string<CharType_t> > &argv) {
@@ -1936,12 +1960,12 @@ long convert_digit(wchar_t d, int base) {
     return res;
 }
 
-// Test if the specified character is in a range that fish uses interally to store special tokens.
-//
-// NOTE: This is used when tokenizing the input. It is also used when reading input, before
-// tokenization, to replace such chars with REPLACEMENT_WCHAR if they're not part of a quoted
-// string. We don't want external input to be able to feed reserved characters into our lexer/parser
-// or code evaluator.
+/// Test if the specified character is in a range that fish uses interally to store special tokens.
+///
+/// NOTE: This is used when tokenizing the input. It is also used when reading input, before
+/// tokenization, to replace such chars with REPLACEMENT_WCHAR if they're not part of a quoted
+/// string. We don't want external input to be able to feed reserved characters into our
+/// lexer/parser or code evaluator.
 //
 // TODO: Actually implement the replacement as documented above.
 bool fish_reserved_codepoint(wchar_t c) {
@@ -1949,3 +1973,50 @@ bool fish_reserved_codepoint(wchar_t c) {
            (c >= ENCODE_DIRECT_BASE && c < ENCODE_DIRECT_END) ||
            (c >= INPUT_COMMON_BASE && c < INPUT_COMMON_END);
 }
+
+/// Reopen stdin, stdout and/or stderr on /dev/null. This is invoked when we find that our tty has
+/// become invalid.
+void redirect_tty_output() {
+    struct termios t;
+    int fd = open("/dev/null", O_WRONLY);
+    if (tcgetattr(STDIN_FILENO, &t) == -1 && errno == EIO) dup2(fd, STDIN_FILENO);
+    if (tcgetattr(STDOUT_FILENO, &t) == -1 && errno == EIO) dup2(fd, STDOUT_FILENO);
+    if (tcgetattr(STDERR_FILENO, &t) == -1 && errno == EIO) dup2(fd, STDERR_FILENO);
+    close(fd);
+}
+
+/// Display a failed assertion message, dump a stack trace if possible, then die.
+[[noreturn]] void __assert(const char *msg, const char *file, size_t line, int error) {
+    if (error) {
+        debug(0, L"%s:%zu: failed assertion: %s: errno %d (%s)", file, line, msg, error,
+              strerror(error));
+    } else {
+        debug(0, L"%s:%zu: failed assertion: %s", file, line, msg);
+    }
+    show_stackframe(L'E', 99, 1);
+    abort();
+}
+
+/// Test if the given char is valid in a variable name.
+bool valid_var_name_char(wchar_t chr) { return fish_iswalnum(chr) || chr == L'_'; }
+
+/// Test if the given string is a valid variable name.
+bool valid_var_name(const wchar_t *str) {
+    if (str[0] == L'\0') return false;
+    while (*str) {
+        if (!valid_var_name_char(*str)) return false;
+        str++;
+    }
+    return true;
+}
+
+/// Test if the given string is a valid variable name.
+bool valid_var_name(const wcstring &str) { return valid_var_name(str.c_str()); }
+
+/// Test if the string is a valid function name.
+bool valid_func_name(const wcstring &str) {
+    if (str.size() == 0) return false;
+    if (str.at(0) == L'-') return false;
+    if (str.find_first_of(L'/') != wcstring::npos) return false;
+    return true;
+}
