@@ -11,11 +11,14 @@
 #include <errno.h>
 #include <getopt.h>
 #include <signal.h>
+#include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <termios.h>
 #include <unistd.h>
 #include <wchar.h>
+
 #include <memory>
 #include <string>
 #include <vector>
@@ -33,10 +36,10 @@
 
 struct config_paths_t determine_config_directory_paths(const char *argv0);
 
-static const char *ctrl_symbolic_names[] = {NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL, "\\a",
-                                            "\\b", "\\t", "\\n", "\\v", "\\f", "\\r", NULL, NULL,
-                                            NULL,  NULL,  NULL,  NULL,  NULL,  NULL,  NULL, NULL,
-                                            NULL,  NULL,  NULL,  "\\e", NULL,  NULL,  NULL, NULL};
+static const wchar_t *ctrl_symbolic_names[] = {
+    NULL,   NULL,   NULL,   NULL, NULL, NULL,   NULL, L"\\a", L"\\b", L"\\t", L"\\n",
+    L"\\v", L"\\f", L"\\r", NULL, NULL, NULL,   NULL, NULL,   NULL,   NULL,   NULL,
+    NULL,   NULL,   NULL,   NULL, NULL, L"\\e", NULL, NULL,   NULL,   NULL};
 static bool keep_running = true;
 
 /// Return true if the recent sequence of characters indicates the user wants to exit the program.
@@ -48,9 +51,17 @@ static bool should_exit(wchar_t wc) {
     recent_chars[1] = recent_chars[2];
     recent_chars[2] = recent_chars[3];
     recent_chars[3] = c;
-    return (memcmp(recent_chars, "exit", 4) == 0 || memcmp(recent_chars, "quit", 4) == 0 ||
-            memcmp(recent_chars + 2, "\x3\x3", 2) == 0 ||  // ctrl-C, ctrl-C
-            memcmp(recent_chars + 2, "\x4\x4", 2) == 0);   // ctrl-D, ctrl-D
+    if (c == shell_modes.c_cc[VINTR]) {
+        if (recent_chars[2] == shell_modes.c_cc[VINTR]) return true;
+        fwprintf(stderr, L"Press [ctrl-%c] again to exit\n", shell_modes.c_cc[VINTR] + 0x40);
+        return false;
+    }
+    if (c == shell_modes.c_cc[VEOF]) {
+        if (recent_chars[2] == shell_modes.c_cc[VEOF]) return true;
+        fwprintf(stderr, L"Press [ctrl-%c] again to exit\n", shell_modes.c_cc[VEOF] + 0x40);
+        return false;
+    }
+    return memcmp(recent_chars, "exit", 4) == 0 || memcmp(recent_chars, "quit", 4) == 0;
 }
 
 /// Return the name if the recent sequence of characters matches a known terminfo sequence.
@@ -79,74 +90,67 @@ static char *sequence_name(wchar_t wc) {
     return NULL;
 }
 
-/// Return true if the character must be escaped in the sequence of chars to be bound in `bind`
-/// command.
+/// Return true if the character must be escaped when used in the sequence of chars to be bound in
+/// a `bind` command.
 static bool must_escape(wchar_t wc) {
-    switch (wc) {
-        case '[':
-        case ']':
-        case '(':
-        case ')':
-        case '<':
-        case '>':
-        case '{':
-        case '}':
-        case '*':
-        case '\\':
-        case '?':
-        case '$':
-        case '#':
-        case ';':
-        case '&':
-        case '|':
-        case '\'':
-        case '"':
-            return true;
-        default:
-            return false;
+    return wcschr(L"[]()<>{}*\\?$#;&|'\"", wc) != NULL;
+}
+
+static void ctrl_to_symbol(wchar_t *buf, int buf_len, wchar_t wc, bool bind_friendly) {
+    if (ctrl_symbolic_names[wc]) {
+        if (bind_friendly) {
+            swprintf(buf, buf_len, L"%ls", ctrl_symbolic_names[wc]);
+        } else {
+            swprintf(buf, buf_len, L"\\c%c  (or %ls)", wc + 0x40, ctrl_symbolic_names[wc]);
+        }
+    } else {
+        swprintf(buf, buf_len, L"\\c%c", wc + 0x40);
     }
 }
 
-static char *char_to_symbol(wchar_t wc, bool bind_friendly) {
-    static char buf[128];
-
-    if (wc < ' ') {
-        // ASCII control character.
-        if (ctrl_symbolic_names[wc]) {
-            if (bind_friendly) {
-                snprintf(buf, sizeof(buf), "%s", ctrl_symbolic_names[wc]);
-            } else {
-                snprintf(buf, sizeof(buf), "\\c%c  (or %s)", wc + 64, ctrl_symbolic_names[wc]);
-            }
-        } else {
-            snprintf(buf, sizeof(buf), "\\c%c", wc + 64);
-        }
-    } else if (wc == ' ') {
-        // The "space" character.
-        if (bind_friendly) {
-            snprintf(buf, sizeof(buf), "\\x%X", wc);
-        } else {
-            snprintf(buf, sizeof(buf), "\\x%X  (aka \"space\")", wc);
-        }
-    } else if (wc == 0x7F) {
-        // The "del" character.
-        if (bind_friendly) {
-            snprintf(buf, sizeof(buf), "\\x%X", wc);
-        } else {
-            snprintf(buf, sizeof(buf), "\\x%X  (aka \"del\")", wc);
-        }
-    } else if (wc < 0x80) {
-        // ASCII characters that are not control characters.
-        if (bind_friendly && must_escape(wc)) {
-            snprintf(buf, sizeof(buf), "\\%c", wc);
-        } else {
-            snprintf(buf, sizeof(buf), "%c", wc);
-        }
-    } else if (wc <= 0xFFFF) {
-        snprintf(buf, sizeof(buf), "\\u%04X", wc);
+static void space_to_symbol(wchar_t *buf, int buf_len, wchar_t wc, bool bind_friendly) {
+    if (bind_friendly) {
+        swprintf(buf, buf_len, L"\\x%X", wc);
     } else {
-        snprintf(buf, sizeof(buf), "\\U%06X", wc);
+        swprintf(buf, buf_len, L"\\x%X  (aka \"space\")", wc);
     }
+}
+
+static void del_to_symbol(wchar_t *buf, int buf_len, wchar_t wc, bool bind_friendly) {
+    if (bind_friendly) {
+        swprintf(buf, buf_len, L"\\x%X", wc);
+    } else {
+        swprintf(buf, buf_len, L"\\x%X  (aka \"del\")", wc);
+    }
+}
+
+static void ascii_printable_to_symbol(wchar_t *buf, int buf_len, wchar_t wc, bool bind_friendly) {
+    if (bind_friendly && must_escape(wc)) {
+        swprintf(buf, buf_len, L"\\%c", wc);
+    } else {
+        swprintf(buf, buf_len, L"%c", wc);
+    }
+}
+
+/// Convert a wide-char to a symbol that can be used in our output. The use of a static buffer
+/// requires that the returned string be used before we are called again.
+static wchar_t *char_to_symbol(wchar_t wc, bool bind_friendly) {
+    static wchar_t buf[64];
+
+    if (wc < L' ') {  // ASCII control character
+        ctrl_to_symbol(buf, sizeof(buf) / sizeof(*buf), wc, bind_friendly);
+    } else if (wc == L' ') {  // the "space" character
+        space_to_symbol(buf, sizeof(buf) / sizeof(*buf), wc, bind_friendly);
+    } else if (wc == 0x7F) {  // the "del" character
+        del_to_symbol(buf, sizeof(buf) / sizeof(*buf), wc, bind_friendly);
+    } else if (wc < 0x80) {  // ASCII characters that are not control characters
+        ascii_printable_to_symbol(buf, sizeof(buf) / sizeof(*buf), wc, bind_friendly);
+    } else if (wc <= 0xFFFF) {  // BMP Unicode chararacter
+        swprintf(buf, sizeof(buf) / sizeof(*buf), L"\\u%04X", wc);
+    } else {  // Non-BMP Unicode chararacter
+        swprintf(buf, sizeof(buf) / sizeof(*buf), L"\\U%06X", wc);
+    }
+
     return buf;
 }
 
@@ -156,23 +160,23 @@ static void add_char_to_bind_command(wchar_t wc, std::vector<wchar_t> &bind_char
 
 static void output_bind_command(std::vector<wchar_t> &bind_chars) {
     if (bind_chars.size()) {
-        fputs("bind ", stdout);
+        fputws(L"bind ", stdout);
         for (size_t i = 0; i < bind_chars.size(); i++) {
-            fputs(char_to_symbol(bind_chars[i], true), stdout);
+            fputws(char_to_symbol(bind_chars[i], true), stdout);
         }
-        fputs(" 'do something'\n", stdout);
+        fputws(L" 'do something'\n", stdout);
         bind_chars.clear();
     }
 }
 
 static void output_info_about_char(wchar_t wc) {
-    fprintf(stderr, "hex: %4X  char: %s\n", wc, char_to_symbol(wc, false));
+    fwprintf(stderr, L"hex: %4X  char: %ls\n", wc, char_to_symbol(wc, false));
 }
 
 static bool output_matching_key_name(wchar_t wc) {
     char *name = sequence_name(wc);
     if (name) {
-        printf("bind -k %s 'do something'\n", name);
+        fwprintf(stdout, L"bind -k %s 'do something'\n", name);
         free(name);
         return true;
     }
@@ -184,11 +188,11 @@ static double output_elapsed_time(double prev_tstamp, bool first_char_seen) {
     double now = timef();
     long long int delta_tstamp_us = 1000000 * (now - prev_tstamp);
 
-    if (delta_tstamp_us >= 200000 && first_char_seen) putc('\n', stderr);
+    if (delta_tstamp_us >= 200000 && first_char_seen) fputwc(L'\n', stderr);
     if (delta_tstamp_us >= 1000000) {
-        fprintf(stderr, "              ");
+        fwprintf(stderr, L"              ");
     } else {
-        fprintf(stderr, "(%3lld.%03lld ms)  ", delta_tstamp_us / 1000, delta_tstamp_us % 1000);
+        fwprintf(stderr, L"(%3lld.%03lld ms)  ", delta_tstamp_us / 1000, delta_tstamp_us % 1000);
     }
     return now;
 }
@@ -199,16 +203,20 @@ static void process_input(bool continuous_mode) {
     double prev_tstamp = 0.0;
     std::vector<wchar_t> bind_chars;
 
-    fprintf(stderr, "Press a key\n\n");
+    fwprintf(stderr, L"Press a key\n\n");
     while (keep_running) {
-        wchar_t wc = input_common_readch(true);
+        wchar_t wc;
+        if (reader_interrupted()) {
+            wc = shell_modes.c_cc[VINTR];
+        } else {
+            wc = input_common_readch(true);
+        }
         if (wc == R_TIMEOUT || wc == R_EOF) {
             output_bind_command(bind_chars);
             if (first_char_seen && !continuous_mode) {
                 return;
-            } else {
-                continue;
             }
+            continue;
         }
 
         prev_tstamp = output_elapsed_time(prev_tstamp, first_char_seen);
@@ -219,7 +227,7 @@ static void process_input(bool continuous_mode) {
         }
 
         if (should_exit(wc)) {
-            fprintf(stderr, "\nExiting at your request.\n");
+            fwprintf(stderr, L"\nExiting at your request.\n");
             break;
         }
 
@@ -231,12 +239,11 @@ static void process_input(bool continuous_mode) {
 /// Otherwise just report receipt of the signal.
 static struct sigaction old_sigactions[32];
 static void signal_handler(int signo, siginfo_t *siginfo, void *siginfo_arg) {
-    debug(2, L"signal #%d (%ls) received", signo, sig2wcs(signo));
-    // SIGINT isn't included in the following conditional because it is handled specially by fish.
-    // Specifically, it causes \cC to be reinserted into the tty input stream.
+    fwprintf(stdout, _(L"signal #%d (%ls) received\n"), signo, sig2wcs(signo));
     if (signo == SIGHUP || signo == SIGTERM || signo == SIGABRT || signo == SIGSEGV) {
         keep_running = false;
     }
+
     if (old_sigactions[signo].sa_handler != SIG_IGN &&
         old_sigactions[signo].sa_handler != SIG_DFL) {
         int needs_siginfo = old_sigactions[signo].sa_flags & SA_SIGINFO;
@@ -278,16 +285,15 @@ static void setup_and_process_keys(bool continuous_mode) {
     setup_fork_guards();
     env_init();
     reader_init();
-    input_init();
     proc_push_interactive(1);
-    signal_set_handlers();
     install_our_signal_handlers();
 
     if (continuous_mode) {
-        fprintf(stderr, "\n");
-        fprintf(stderr, "To terminate this program type \"exit\" or \"quit\" in this window,\n");
-        fprintf(stderr, "or press [ctrl-C] or [ctrl-D] twice in a row.\n");
-        fprintf(stderr, "\n");
+        fwprintf(stderr, L"\n");
+        fwprintf(stderr, L"To terminate this program type \"exit\" or \"quit\" in this window,\n");
+        fwprintf(stderr, L"or press [ctrl-%c] or [ctrl-%c] twice in a row.\n",
+                 shell_modes.c_cc[VINTR] + 0x40, shell_modes.c_cc[VEOF] + 0x40);
+        fwprintf(stderr, L"\n");
     }
 
     process_input(continuous_mode);
@@ -297,9 +303,36 @@ static void setup_and_process_keys(bool continuous_mode) {
     reader_destroy();
 }
 
-int main(int argc, char **argv) {
-    program_name = L"fish_key_reader";
-    bool continuous_mode = false;
+static bool parse_debug_level_flag() {
+    errno = 0;
+    char *end;
+    long tmp = strtol(optarg, &end, 10);
+
+    if (tmp >= 0 && tmp <= 10 && !*end && !errno) {
+        debug_level = (int)tmp;
+    } else {
+        fwprintf(stderr, _(L"Invalid value '%s' for debug-level flag\n"), optarg);
+        return false;
+    }
+
+    return true;
+}
+
+static bool parse_debug_frames_flag() {
+    errno = 0;
+    char *end;
+    long tmp = strtol(optarg, &end, 10);
+    if (tmp > 0 && tmp <= 128 && !*end && !errno) {
+        debug_stack_frames = (int)tmp;
+    } else {
+        fwprintf(stderr, _(L"Invalid value '%s' for debug-stack-frames flag\n"), optarg);
+        return false;
+    }
+
+    return true;
+}
+
+static bool parse_flags(int argc, char **argv, bool *continuous_mode) {
     const char *short_opts = "+cd:D:h";
     const struct option long_opts[] = {{"continuous", no_argument, NULL, 'c'},
                                        {"debug-level", required_argument, NULL, 'd'},
@@ -307,65 +340,53 @@ int main(int argc, char **argv) {
                                        {"help", no_argument, NULL, 'h'},
                                        {NULL, 0, NULL, 0}};
     int opt;
-    while ((opt = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
+    bool error = false;
+    while (!error && (opt = getopt_long(argc, argv, short_opts, long_opts, NULL)) != -1) {
         switch (opt) {
-            case 0: {
-                fprintf(stderr, "getopt_long() unexpectedly returned zero\n");
-                exit(1);
-            }
             case 'c': {
-                continuous_mode = true;
+                *continuous_mode = true;
                 break;
             }
             case 'h': {
                 print_help("fish_key_reader", 0);
-                exit(0);
+                error = true;
+                break;
             }
             case 'd': {
-                char *end;
-                long tmp;
-
-                errno = 0;
-                tmp = strtol(optarg, &end, 10);
-
-                if (tmp >= 0 && tmp <= 10 && !*end && !errno) {
-                    debug_level = (int)tmp;
-                } else {
-                    fwprintf(stderr, _(L"Invalid value '%s' for debug-level flag"), optarg);
-                    exit(1);
-                }
+                error = !parse_debug_level_flag();
                 break;
             }
             case 'D': {
-                char *end;
-                long tmp;
-
-                errno = 0;
-                tmp = strtol(optarg, &end, 10);
-
-                if (tmp > 0 && tmp <= 128 && !*end && !errno) {
-                    debug_stack_frames = (int)tmp;
-                } else {
-                    fwprintf(stderr, _(L"Invalid value '%s' for debug-stack-frames flag"), optarg);
-                    exit(1);
-                }
+                error = !parse_debug_frames_flag();
                 break;
             }
             default: {
                 // We assume getopt_long() has already emitted a diagnostic msg.
-                exit(1);
+                error = true;
+                break;
             }
         }
     }
 
+    if (error) return false;
+
     argc -= optind;
     if (argc != 0) {
-        fprintf(stderr, "Expected no arguments, got %d\n", argc);
-        return 1;
+        fwprintf(stderr, L"Expected no arguments, got %d\n", argc);
+        return false;
     }
 
+    return true;
+}
+
+int main(int argc, char **argv) {
+    program_name = L"fish_key_reader";
+    bool continuous_mode = false;
+
+    if (!parse_flags(argc, argv, &continuous_mode)) return 1;
+
     if (!isatty(STDIN_FILENO)) {
-        fprintf(stderr, "Stdin must be attached to a tty.\n");
+        fwprintf(stderr, L"Stdin must be attached to a tty.\n");
         return 1;
     }
 
