@@ -2,22 +2,15 @@
 #include "config.h"
 
 #include <errno.h>
-#include <unistd.h>
+#include <termios.h>
 #include <wchar.h>
-#if HAVE_NCURSES_H
-#include <ncurses.h>
-#elif HAVE_NCURSES_CURSES_H
-#include <ncurses/curses.h>
-#else
-#include <curses.h>
-#endif
+#include <wctype.h>
 #if HAVE_TERM_H
 #include <term.h>
 #elif HAVE_NCURSES_TERM_H
 #include <ncurses/term.h>
 #endif
-#include <stdlib.h>
-#include <wctype.h>
+
 #include <algorithm>
 #include <memory>
 #include <string>
@@ -30,14 +23,12 @@
 #include "input.h"
 #include "input_common.h"
 #include "io.h"
-#include "output.h"
 #include "parser.h"
 #include "proc.h"
 #include "reader.h"
 #include "signal.h"  // IWYU pragma: keep
 #include "wutil.h"   // IWYU pragma: keep
 
-#define DEFAULT_TERM L"ansi"
 #define MAX_INPUT_FUNCTION_ARGS 20
 
 /// Struct representing a keybinding. Returned by input_get_mappings.
@@ -53,11 +44,11 @@ struct input_mapping_t {
     /// New mode that should be switched to after command evaluation.
     wcstring sets_mode;
 
-    input_mapping_t(const wcstring &s, const std::vector<wcstring> &c,
-                    const wcstring &m = DEFAULT_BIND_MODE, const wcstring &sm = DEFAULT_BIND_MODE)
+    input_mapping_t(const wcstring &s, const std::vector<wcstring> &c, const wcstring &m,
+                    const wcstring &sm)
         : seq(s), commands(c), mode(m), sets_mode(sm) {
-        static unsigned int s_last_input_mapping_specification_order = 0;
-        specification_order = ++s_last_input_mapping_specification_order;
+        static unsigned int s_last_input_map_spec_order = 0;
+        specification_order = ++s_last_input_map_spec_order;
     }
 };
 
@@ -198,11 +189,11 @@ static std::vector<terminfo_mapping_t> terminfo_mappings;
 /// List of all terminfo mappings.
 static std::vector<terminfo_mapping_t> mappings;
 
-/// Set to one when the input subsytem has been initialized.
-static bool is_init = false;
+/// Set to true when the input subsytem has been initialized.
+bool input_initialized = false;
 
 /// Initialize terminfo.
-static void input_terminfo_init();
+static void init_input_terminfo();
 
 static wchar_t input_function_args[MAX_INPUT_FUNCTION_ARGS];
 static bool input_function_status;
@@ -217,20 +208,16 @@ wcstring input_get_bind_mode() {
 /// Set the current bind mode.
 void input_set_bind_mode(const wcstring &bm) {
     // Only set this if it differs to not execute variable handlers all the time.
+    // modes may not be empty - empty is a sentinel value meaning to not change the mode
+    assert(!bm.empty());
     if (input_get_bind_mode() != bm.c_str()) {
         env_set(FISH_BIND_MODE_VAR, bm.c_str(), ENV_GLOBAL);
     }
 }
 
 /// Returns the arity of a given input function.
-int input_function_arity(int function) {
-    switch (function) {
-        case R_FORWARD_JUMP:
-        case R_BACKWARD_JUMP: {
-            return 1;
-        }
-        default: { return 0; }
-    }
+static int input_function_arity(int function) {
+    return (function == R_FORWARD_JUMP || function == R_BACKWARD_JUMP) ? 1 : 0;
 }
 
 /// Sets the return status of the most recently executed input function.
@@ -261,8 +248,9 @@ void input_mapping_add(const wchar_t *sequence, const wchar_t *const *commands,
     CHECK(mode, );
     CHECK(sets_mode, );
 
-    // debug( 0, L"Add mapping from %ls to %ls in mode %ls", escape(sequence, ESCAPE_ALL).c_str(),
-    // escape(command, ESCAPE_ALL).c_str(), mode);
+    // debug( 0, L"Add mapping from %ls to %ls in mode %ls", escape_string(sequence,
+    // ESCAPE_ALL).c_str(),
+    // escape_string(command, ESCAPE_ALL).c_str(), mode);
 
     // Remove existing mappings with this sequence.
     const wcstring_list_t commands_vector(commands, commands + commands_len);
@@ -295,94 +283,17 @@ static int interrupt_handler() {
     if (job_reap(1)) reader_repaint_needed();
     // Tell the reader an event occured.
     if (reader_reading_interrupted()) {
-        // Return 3, i.e. the character read by a Control-C.
-        return 3;
+        return shell_modes.c_cc[VINTR];
     }
 
     return R_NULL;
 }
 
-void update_fish_color_support(void) {
-    // Detect or infer term256 support. If fish_term256 is set, we respect it;
-    // otherwise infer it from the TERM variable or use terminfo.
-    env_var_t fish_term256 = env_get_string(L"fish_term256");
-    env_var_t term = env_get_string(L"TERM");
-    bool support_term256 = false;  // default to no support
-    if (!fish_term256.missing_or_empty()) {
-        support_term256 = from_string<bool>(fish_term256);
-        debug(2, L"256 color support determined by 'fish_term256'");
-    } else if (term.find(L"256color") != wcstring::npos) {
-        // TERM=*256color*: Explicitly supported.
-        support_term256 = true;
-        debug(2, L"256 color support enabled for '256color' in TERM");
-    } else if (term.find(L"xterm") != wcstring::npos) {
-        // Assume that all xterms are 256, except for OS X SnowLeopard
-        const env_var_t prog = env_get_string(L"TERM_PROGRAM");
-        const env_var_t progver = env_get_string(L"TERM_PROGRAM_VERSION");
-        if (prog == L"Apple_Terminal" && !progver.missing_or_empty()) {
-            // OS X Lion is version 300+, it has 256 color support
-            if (strtod(wcs2str(progver), NULL) > 300) {
-                support_term256 = true;
-                debug(2, L"256 color support enabled for TERM=xterm + modern Terminal.app");
-            }
-        } else {
-            support_term256 = true;
-            debug(2, L"256 color support enabled for TERM=xterm");
-        }
-    } else if (cur_term != NULL) {
-        // See if terminfo happens to identify 256 colors
-        support_term256 = (max_colors >= 256);
-        debug(2, L"256 color support: using %d colors per terminfo", max_colors);
-    } else {
-        debug(2, L"256 color support not enabled (yet)");
-    }
-
-    env_var_t fish_term24bit = env_get_string(L"fish_term24bit");
-    bool support_term24bit;
-    if (!fish_term24bit.missing_or_empty()) {
-        support_term24bit = from_string<bool>(fish_term24bit);
-        debug(2, L"'fish_term24bit' preference: 24-bit color %s",
-              support_term24bit ? L"enabled" : L"disabled");
-    } else {
-        // We don't attempt to infer term24 bit support yet.
-        support_term24bit = false;
-    }
-
-    color_support_t support = (support_term256 ? color_support_term256 : 0) |
-                              (support_term24bit ? color_support_term24bit : 0);
-    output_set_color_support(support);
-}
-
-int input_init() {
-    if (is_init) return 1;
-    is_init = true;
+/// Set up arrays used by readch to detect escape sequences for special keys and perform related
+/// initializations for our input subsystem.
+void init_input() {
     input_common_init(&interrupt_handler);
-
-    int err_ret;
-    if (setupterm(NULL, STDOUT_FILENO, &err_ret) == ERR) {
-        debug(0, _(L"Could not set up terminal"));
-        env_var_t term = env_get_string(L"TERM");
-        if (term.missing_or_empty()) {
-            debug(0, _(L"TERM environment variable not set"));
-        } else {
-            debug(0, _(L"TERM environment variable set to '%ls'"), term.c_str());
-            debug(0, _(L"Check that this terminal type is supported on this system"));
-        }
-
-        env_set(L"TERM", DEFAULT_TERM, ENV_GLOBAL | ENV_EXPORT);
-        if (setupterm(NULL, STDOUT_FILENO, &err_ret) == ERR) {
-            debug(0,
-                  _(L"Could not set up terminal using the fallback terminal type '%ls' - exiting"),
-                  DEFAULT_TERM);
-            exit_without_destructors(1);
-        } else {
-            debug(0, _(L"Using fallback terminal type '%ls'"), DEFAULT_TERM);
-        }
-    } else {
-    }
-
-    input_terminfo_init();
-    update_fish_color_support();
+    init_input_terminfo();
 
     // If we have no keybindings, add a few simple defaults.
     if (mapping_list.empty()) {
@@ -395,17 +306,14 @@ int input_init() {
         input_mapping_add(L"\x5", L"bind");
     }
 
-    return 1;
+    input_initialized = true;
+    return;
 }
 
 void input_destroy() {
-    if (!is_init) return;
-    is_init = false;
+    if (!input_initialized) return;
+    input_initialized = false;
     input_common_destroy();
-
-    if (del_curterm(cur_term) == ERR) {
-        debug(0, _(L"Error while closing terminfo"));
-    }
 }
 
 void input_function_push_arg(wchar_t arg) {
@@ -453,7 +361,7 @@ static void input_mapping_execute(const input_mapping_t &m, bool allow_commands)
 
     // !has_functions && !has_commands: only set bind mode
     if (!has_commands && !has_functions) {
-        input_set_bind_mode(m.sets_mode);
+        if (!m.sets_mode.empty()) input_set_bind_mode(m.sets_mode);
         return;
     }
 
@@ -492,15 +400,16 @@ static void input_mapping_execute(const input_mapping_t &m, bool allow_commands)
         input_common_next_ch(R_NULL);
     }
 
-    input_set_bind_mode(m.sets_mode);
+    // Empty bind mode indicates to not reset the mode (#2871)
+    if (!m.sets_mode.empty()) input_set_bind_mode(m.sets_mode);
 }
 
 /// Try reading the specified function mapping.
 static bool input_mapping_is_match(const input_mapping_t &m) {
-    wint_t c = 0;
+    wchar_t c = 0;
     int j;
 
-    debug(2, L"trying to match mapping %ls", escape(m.seq.c_str(), ESCAPE_ALL).c_str());
+    debug(4, L"trying to match mapping %ls", escape_string(m.seq.c_str(), ESCAPE_ALL).c_str());
     const wchar_t *str = m.seq.c_str();
     for (j = 0; str[j] != L'\0'; j++) {
         bool timed = (j > 0 && iswcntrl(str[0]));
@@ -512,7 +421,8 @@ static bool input_mapping_is_match(const input_mapping_t &m) {
     }
 
     if (str[j] == L'\0') {
-        // debug(0, L"matched mapping %ls (%ls)\n", escape(m.seq.c_str(), ESCAPE_ALL).c_str(),
+        // debug(0, L"matched mapping %ls (%ls)\n", escape_string(m.seq.c_str(),
+        // ESCAPE_ALL).c_str(),
         // m.command.c_str());
         // We matched the entire sequence.
         return true;
@@ -538,7 +448,8 @@ static void input_mapping_execute_matching_or_generic(bool allow_commands) {
     for (size_t i = 0; i < mapping_list.size(); i++) {
         const input_mapping_t &m = mapping_list.at(i);
 
-        // debug(0, L"trying mapping (%ls,%ls,%ls)\n", escape(m.seq.c_str(), ESCAPE_ALL).c_str(),
+        // debug(0, L"trying mapping (%ls,%ls,%ls)\n", escape_string(m.seq.c_str(),
+        // ESCAPE_ALL).c_str(),
         //           m.mode.c_str(), m.sets_mode.c_str());
 
         if (m.mode != bind_mode) {
@@ -614,8 +525,9 @@ wint_t input_readch(bool allow_commands) {
                     if (input_function_status) {
                         return input_readch();
                     }
-                    while ((c = input_common_readch(0)) && c >= R_MIN && c <= R_MAX) {
-                        // do nothing
+                    c = input_common_readch(0);
+                    while (c >= R_MIN && c <= R_MAX) {
+                        c = input_common_readch(0);
                     }
                     input_common_next_ch(c);
                     return input_readch();
@@ -678,8 +590,10 @@ bool input_mapping_get(const wcstring &sequence, const wcstring &mode, wcstring_
     return result;
 }
 
-/// Add all terminfo mappings.
-static void input_terminfo_init() {
+/// Add all terminfo mappings and cache other terminfo facts we care about.
+static void init_input_terminfo() {
+    assert(curses_initialized);
+    if (!cur_term) return;  // setupterm() failed so we can't referency any key definitions
     const terminfo_mapping_t tinfos[] = {
         TERMINFO_ADD(key_a1),
         TERMINFO_ADD(key_a3),
@@ -843,13 +757,12 @@ static void input_terminfo_init() {
 
 bool input_terminfo_get_sequence(const wchar_t *name, wcstring *out_seq) {
     ASSERT_IS_MAIN_THREAD();
+    assert(input_initialized);
+    CHECK(name, 0);
 
     const char *res = 0;
     int err = ENOENT;
 
-    CHECK(name, 0);
-    input_init();
-
     for (size_t i = 0; i < terminfo_mappings.size(); i++) {
         const terminfo_mapping_t &m = terminfo_mappings.at(i);
         if (!wcscmp(name, m.name)) {
@@ -869,7 +782,7 @@ bool input_terminfo_get_sequence(const wchar_t *name, wcstring *out_seq) {
 }
 
 bool input_terminfo_get_name(const wcstring &seq, wcstring *out_name) {
-    input_init();
+    assert(input_initialized);
 
     for (size_t i = 0; i < terminfo_mappings.size(); i++) {
         terminfo_mapping_t &m = terminfo_mappings.at(i);
@@ -889,11 +802,10 @@ bool input_terminfo_get_name(const wcstring &seq, wcstring *out_name) {
 }
 
 wcstring_list_t input_terminfo_get_names(bool skip_null) {
+    assert(input_initialized);
     wcstring_list_t result;
     result.reserve(terminfo_mappings.size());
 
-    input_init();
-
     for (size_t i = 0; i < terminfo_mappings.size(); i++) {
         terminfo_mapping_t &m = terminfo_mappings.at(i);
 
