@@ -8,7 +8,6 @@
 // for the execution to finish to see them.
 #include "config.h"  // IWYU pragma: keep
 
-#include <assert.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -21,6 +20,7 @@
 #include <algorithm>
 #include <memory>
 #include <string>
+#include <type_traits>
 #include <vector>
 
 #include "builtin.h"
@@ -75,10 +75,9 @@ static wcstring profiling_cmd_name_for_redirectable_block(const parse_node_t &no
     return result;
 }
 
-parse_execution_context_t::parse_execution_context_t(moved_ref<parse_node_tree_t> t,
-                                                     const wcstring &s, parser_t *p,
-                                                     int initial_eval_level)
-    : tree(t),
+parse_execution_context_t::parse_execution_context_t(parse_node_tree_t t, const wcstring &s,
+                                                     parser_t *p, int initial_eval_level)
+    : tree(std::move(t)),
       src(s),
       parser(p),
       eval_level(initial_eval_level),
@@ -103,7 +102,6 @@ node_offset_t parse_execution_context_t::get_offset(const parse_node_t &node) co
     const parse_node_t *addr = &node;
     const parse_node_t *base = &this->tree.at(0);
     assert(addr >= base);
-    assert(addr - base < SOURCE_OFFSET_INVALID);
     node_offset_t offset = static_cast<node_offset_t>(addr - base);
     assert(offset < this->tree.size());
     assert(&tree.at(offset) == &node);
@@ -220,13 +218,12 @@ parse_execution_context_t::cancellation_reason(const block_t *block) const {
     if (parser && parser->cancellation_requested) {
         return execution_cancellation_skip;
     }
-    if (block && block->loop_status != LOOP_NORMAL) {
-        // Nasty hack - break and continue set the 'skip' flag as well as the loop status flag.
-        return execution_cancellation_loop_control;
-    }
     if (block && block->skip) {
         return execution_cancellation_skip;
     }
+    if (block && block->loop_status != LOOP_NORMAL) {
+        return execution_cancellation_loop_control;
+    }
     return execution_cancellation_none;
 }
 
@@ -267,9 +264,8 @@ parse_execution_result_t parse_execution_context_t::run_if_statement(
     assert(statement.type == symbol_if_statement);
 
     // Push an if block.
-    if_block_t *ib = new if_block_t();
+    if_block_t *ib = parser->push_block<if_block_t>();
     ib->node_offset = this->get_offset(statement);
-    parser->push_block(ib);
 
     parse_execution_result_t result = parse_execution_success;
 
@@ -306,7 +302,7 @@ parse_execution_result_t parse_execution_context_t::run_if_statement(
         } else if (else_clause->child_count == 0) {
             // 'if' condition failed, no else clause, return 0, we're done.
             job_list_to_execute = NULL;
-            proc_set_last_status(STATUS_BUILTIN_OK);
+            proc_set_last_status(STATUS_CMD_OK);
             break;
         } else {
             // We have an 'else continuation' (either else-if or else).
@@ -329,7 +325,7 @@ parse_execution_result_t parse_execution_context_t::run_if_statement(
         run_job_list(*job_list_to_execute, ib);
     } else {
         // No job list means no sucessful conditions, so return 0 (issue #1443).
-        proc_set_last_status(STATUS_BUILTIN_OK);
+        proc_set_last_status(STATUS_CMD_OK);
     }
 
     // It's possible there's a last-minute cancellation (issue #1297).
@@ -349,14 +345,9 @@ parse_execution_result_t parse_execution_context_t::run_begin_statement(
     assert(header.type == symbol_begin_header);
     assert(contents.type == symbol_job_list);
 
-    // Basic begin/end block. Push a scope block.
-    scope_block_t *sb = new scope_block_t(BEGIN);
-    parser->push_block(sb);
-
-    // Run the job list.
+    // Basic begin/end block. Push a scope block, run jobs, pop it
+    scope_block_t *sb = parser->push_block<scope_block_t>(BEGIN);
     parse_execution_result_t ret = run_job_list(contents, sb);
-
-    // Pop the block.
     parser->pop_block(sb);
 
     return ret;
@@ -372,38 +363,41 @@ parse_execution_result_t parse_execution_context_t::run_function_statement(
     wcstring_list_t argument_list;
     parse_execution_result_t result = this->determine_arguments(header, &argument_list, failglob);
 
-    if (result == parse_execution_success) {
-        // The function definition extends from the end of the header to the function end. It's not
-        // just the range of the contents because that loses comments - see issue #1710.
-        assert(block_end_command.has_source());
-        size_t contents_start = header.source_start + header.source_length;
-        size_t contents_end =
-            block_end_command.source_start;  // 1 past the last character in the function definition
-        assert(contents_end >= contents_start);
-
-        // Swallow whitespace at both ends.
-        while (contents_start < contents_end && iswspace(this->src.at(contents_start))) {
-            contents_start++;
-        }
-        while (contents_start < contents_end && iswspace(this->src.at(contents_end - 1))) {
-            contents_end--;
-        }
-
-        assert(contents_end >= contents_start);
-        const wcstring contents_str =
-            wcstring(this->src, contents_start, contents_end - contents_start);
-        int definition_line_offset = this->line_offset_of_character_at_offset(contents_start);
-        wcstring error_str;
-        io_streams_t streams;
-        int err = builtin_function(*parser, streams, argument_list, contents_str,
-                                   definition_line_offset, &error_str);
-        proc_set_last_status(err);
-
-        if (!error_str.empty()) {
-            this->report_error(header, L"%ls", error_str.c_str());
-            result = parse_execution_errored;
-        }
+    if (result != parse_execution_success) {
+        return result;
     }
+
+    // The function definition extends from the end of the header to the function end. It's not
+    // just the range of the contents because that loses comments - see issue #1710.
+    assert(block_end_command.has_source());
+    size_t contents_start = header.source_start + header.source_length;
+    size_t contents_end =
+        block_end_command.source_start;  // 1 past the last character in the function definition
+    assert(contents_end >= contents_start);
+
+    // Swallow whitespace at both ends.
+    while (contents_start < contents_end && iswspace(this->src.at(contents_start))) {
+        contents_start++;
+    }
+    while (contents_start < contents_end && iswspace(this->src.at(contents_end - 1))) {
+        contents_end--;
+    }
+
+    assert(contents_end >= contents_start);
+    const wcstring contents_str =
+        wcstring(this->src, contents_start, contents_end - contents_start);
+    int definition_line_offset = this->line_offset_of_character_at_offset(contents_start);
+    wcstring error_str;
+    io_streams_t streams;
+    int err = builtin_function(*parser, streams, argument_list, contents_str,
+                               definition_line_offset, &error_str);
+    proc_set_last_status(err);
+
+    if (!error_str.empty()) {
+        this->report_error(header, L"%ls", error_str.c_str());
+        result = parse_execution_errored;
+    }
+
     return result;
 }
 
@@ -438,7 +432,7 @@ parse_execution_result_t parse_execution_context_t::run_block_statement(
             break;
         }
         default: {
-            fprintf(stderr, "Unexpected block header: %ls\n", header.describe().c_str());
+            debug(0, L"Unexpected block header: %ls\n", header.describe().c_str());
             PARSER_DIE();
             break;
         }
@@ -468,8 +462,7 @@ parse_execution_result_t parse_execution_context_t::run_for_statement(
         return ret;
     }
 
-    for_block_t *fb = new for_block_t();
-    parser->push_block(fb);
+    for_block_t *fb = parser->push_block<for_block_t>();
 
     // Now drive the for loop.
     const size_t arg_count = argument_sequence.size();
@@ -482,7 +475,6 @@ parse_execution_result_t parse_execution_context_t::run_for_statement(
         const wcstring &val = argument_sequence.at(i);
         env_set(for_var_name, val.c_str(), ENV_LOCAL);
         fb->loop_status = LOOP_NORMAL;
-        fb->skip = 0;
 
         this->run_job_list(block_contents, fb);
 
@@ -491,7 +483,6 @@ parse_execution_result_t parse_execution_context_t::run_for_statement(
             if (fb->loop_status == LOOP_CONTINUE) {
                 // Reset the loop state.
                 fb->loop_status = LOOP_NORMAL;
-                fb->skip = false;
                 continue;
             } else if (fb->loop_status == LOOP_BREAK) {
                 break;
@@ -534,6 +525,10 @@ parse_execution_result_t parse_execution_context_t::run_switch_statement(
         case EXPAND_OK: {
             break;
         }
+        default: {
+            DIE("unexpected expand_string() return value");
+            break;
+        }
     }
 
     if (result == parse_execution_success && switch_values_expanded.size() != 1) {
@@ -542,66 +537,66 @@ parse_execution_result_t parse_execution_context_t::run_switch_statement(
                          switch_values_expanded.size());
     }
 
-    if (result == parse_execution_success) {
-        const wcstring &switch_value_expanded = switch_values_expanded.at(0).completion;
+    if (result != parse_execution_success) {
+        return result;
+    }
 
-        switch_block_t *sb = new switch_block_t();
-        parser->push_block(sb);
+    const wcstring &switch_value_expanded = switch_values_expanded.at(0).completion;
 
-        // Expand case statements.
-        const parse_node_t *case_item_list = get_child(statement, 3, symbol_case_item_list);
+    switch_block_t *sb = parser->push_block<switch_block_t>();
 
-        // Loop while we don't have a match but do have more of the list.
-        const parse_node_t *matching_case_item = NULL;
-        while (matching_case_item == NULL && case_item_list != NULL) {
-            if (should_cancel_execution(sb)) {
-                result = parse_execution_cancelled;
-                break;
-            }
+    // Expand case statements.
+    const parse_node_t *case_item_list = get_child(statement, 3, symbol_case_item_list);
 
-            // Get the next item and the remainder of the list.
-            const parse_node_t *case_item =
-                tree.next_node_in_node_list(*case_item_list, symbol_case_item, &case_item_list);
-            if (case_item == NULL) {
-                // No more items.
-                break;
-            }
+    // Loop while we don't have a match but do have more of the list.
+    const parse_node_t *matching_case_item = NULL;
+    while (matching_case_item == NULL && case_item_list != NULL) {
+        if (should_cancel_execution(sb)) {
+            result = parse_execution_cancelled;
+            break;
+        }
 
-            // Pull out the argument list.
-            const parse_node_t &arg_list = *get_child(*case_item, 1, symbol_argument_list);
+        // Get the next item and the remainder of the list.
+        const parse_node_t *case_item =
+            tree.next_node_in_node_list(*case_item_list, symbol_case_item, &case_item_list);
+        if (case_item == NULL) {
+            // No more items.
+            break;
+        }
 
-            // Expand arguments. A case item list may have a wildcard that fails to expand to
-            // anything. We also report case errors, but don't stop execution; i.e. a case item that
-            // contains an unexpandable process will report and then fail to match.
-            wcstring_list_t case_args;
-            parse_execution_result_t case_result =
-                this->determine_arguments(arg_list, &case_args, failglob);
-            if (case_result == parse_execution_success) {
-                for (size_t i = 0; i < case_args.size(); i++) {
-                    const wcstring &arg = case_args.at(i);
+        // Pull out the argument list.
+        const parse_node_t &arg_list = *get_child(*case_item, 1, symbol_argument_list);
 
-                    // Unescape wildcards so they can be expanded again.
-                    wcstring unescaped_arg = parse_util_unescape_wildcards(arg);
-                    bool match = wildcard_match(switch_value_expanded, unescaped_arg);
+        // Expand arguments. A case item list may have a wildcard that fails to expand to
+        // anything. We also report case errors, but don't stop execution; i.e. a case item that
+        // contains an unexpandable process will report and then fail to match.
+        wcstring_list_t case_args;
+        parse_execution_result_t case_result =
+            this->determine_arguments(arg_list, &case_args, failglob);
+        if (case_result == parse_execution_success) {
+            for (size_t i = 0; i < case_args.size(); i++) {
+                const wcstring &arg = case_args.at(i);
 
-                    // If this matched, we're done.
-                    if (match) {
-                        matching_case_item = case_item;
-                        break;
-                    }
+                // Unescape wildcards so they can be expanded again.
+                wcstring unescaped_arg = parse_util_unescape_wildcards(arg);
+                bool match = wildcard_match(switch_value_expanded, unescaped_arg);
+
+                // If this matched, we're done.
+                if (match) {
+                    matching_case_item = case_item;
+                    break;
                 }
             }
         }
-
-        if (result == parse_execution_success && matching_case_item != NULL) {
-            // Success, evaluate the job list.
-            const parse_node_t *job_list = get_child(*matching_case_item, 3, symbol_job_list);
-            result = this->run_job_list(*job_list, sb);
-        }
-
-        parser->pop_block(sb);
     }
 
+    if (result == parse_execution_success && matching_case_item != NULL) {
+        // Success, evaluate the job list.
+        const parse_node_t *job_list = get_child(*matching_case_item, 3, symbol_job_list);
+        result = this->run_job_list(*job_list, sb);
+    }
+
+    parser->pop_block(sb);
     return result;
 }
 
@@ -611,9 +606,8 @@ parse_execution_result_t parse_execution_context_t::run_while_statement(
     assert(block_contents.type == symbol_job_list);
 
     // Push a while block.
-    while_block_t *wb = new while_block_t();
+    while_block_t *wb = parser->push_block<while_block_t>();
     wb->node_offset = this->get_offset(header);
-    parser->push_block(wb);
 
     parse_execution_result_t ret = parse_execution_success;
 
@@ -648,7 +642,6 @@ parse_execution_result_t parse_execution_context_t::run_while_statement(
             if (wb->loop_status == LOOP_CONTINUE) {
                 // Reset the loop state.
                 wb->loop_status = LOOP_NORMAL;
-                wb->skip = false;
                 continue;
             } else if (wb->loop_status == LOOP_BREAK) {
                 break;
@@ -662,9 +655,8 @@ parse_execution_result_t parse_execution_context_t::run_while_statement(
         }
     }
 
-    /* Done */
+    // Done
     parser->pop_block(wb);
-
     return ret;
 }
 
@@ -692,7 +684,7 @@ parse_execution_result_t parse_execution_context_t::report_errors(
     const parse_error_list_t &error_list) const {
     if (!parser->cancellation_requested) {
         if (error_list.empty()) {
-            fprintf(stderr, "Bug: Error reported but no error text found.");
+            debug(0, "Error reported but no error text found.");
         }
 
         // Get a backtrace.
@@ -700,7 +692,9 @@ parse_execution_result_t parse_execution_context_t::report_errors(
         parser->get_backtrace(src, error_list, &backtrace_and_desc);
 
         // Print it.
-        fprintf(stderr, "%ls", backtrace_and_desc.c_str());
+        if (!should_suppress_stderr_for_tests()) {
+            fwprintf(stderr, L"%ls", backtrace_and_desc.c_str());
+        }
     }
     return parse_execution_errored;
 }
@@ -715,25 +709,26 @@ parse_execution_result_t parse_execution_context_t::report_unmatched_wildcard_er
 
 // Given a command string that might contain fish special tokens return a string without those
 // tokens.
-static wcstring reconstruct_orig_cmd(wcstring cmd_str) {
-    // TODO(krader1961): Figure out what VARIABLE_EXPAND means in this context. After looking at the
-    // code and doing various tests I couldn't figure out why that token would be present when this
-    // code is run. I was therefore unable to determine how to substitute its presence in the error
-    // message.
-    wcstring orig_cmd = cmd_str;
+//
+// TODO(krader1961): Figure out what VARIABLE_EXPAND means in this context. After looking at the
+// code and doing various tests I couldn't figure out why that token would be present when this
+// code is run. I was therefore unable to determine how to substitute its presence in the error
+// message.
+static wcstring reconstruct_orig_str(wcstring tokenized_str) {
+    wcstring orig_str = tokenized_str;
 
-    if (cmd_str.find(VARIABLE_EXPAND_SINGLE) != std::string::npos) {
+    if (tokenized_str.find(VARIABLE_EXPAND_SINGLE) != std::string::npos) {
         // Variable was quoted to force expansion of multiple elements into a single element.
         //
         // The following isn't entirely correct. For example, $abc"$def" will become "$abc$def".
         // However, anyone writing the former is asking for trouble so I don't feel bad about not
         // accurately reconstructing what they typed.
-        wcstring new_cmd_str = wcstring(cmd_str);
-        std::replace(new_cmd_str.begin(), new_cmd_str.end(), (wchar_t)VARIABLE_EXPAND_SINGLE, L'$');
-        orig_cmd = L"\"" + new_cmd_str + L"\"";
+        wcstring new_str = wcstring(tokenized_str);
+        std::replace(new_str.begin(), new_str.end(), (wchar_t)VARIABLE_EXPAND_SINGLE, L'$');
+        orig_str = L"\"" + new_str + L"\"";
     }
 
-    return orig_cmd;
+    return orig_str;
 }
 
 /// Handle the case of command not found.
@@ -766,15 +761,17 @@ parse_execution_result_t parse_execution_context_t::handle_command_not_found(
                                name_str.c_str(), val_str.c_str(), argument.c_str(),
                                ellipsis_str.c_str());
         } else {
+            wcstring assigned_val = reconstruct_orig_str(val_str);
             this->report_error(statement_node, ERROR_BAD_COMMAND_ASSIGN_ERR_MSG, name_str.c_str(),
-                               val_str.c_str());
+                               assigned_val.c_str());
         }
     } else if (wcschr(cmd, L'$') || wcschr(cmd, VARIABLE_EXPAND_SINGLE) ||
                wcschr(cmd, VARIABLE_EXPAND)) {
-        wcstring eval_cmd = reconstruct_orig_cmd(cmd_str);
-        this->report_error(statement_node, _(L"Variables may not be used as commands. In fish, "
-                                             L"please define a function or use 'eval %ls'."),
-                           eval_cmd.c_str());
+        const wchar_t *msg =
+            _(L"Variables may not be used as commands. In fish, "
+              L"please define a function or use 'eval %ls'.");
+        wcstring eval_cmd = reconstruct_orig_str(cmd_str);
+        this->report_error(statement_node, msg, eval_cmd.c_str());
     } else if (err_code != ENOENT) {
         this->report_error(statement_node, _(L"The file '%ls' is not executable by this user"),
                            cmd ? cmd : L"UNKNOWN");
@@ -800,7 +797,7 @@ parse_execution_result_t parse_execution_context_t::handle_command_not_found(
     }
 
     // Set the last proc status appropriately.
-    proc_set_last_status(err_code == ENOENT ? STATUS_UNKNOWN_COMMAND : STATUS_NOT_EXECUTABLE);
+    proc_set_last_status(err_code == ENOENT ? STATUS_CMD_UNKNOWN : STATUS_NOT_EXECUTABLE);
 
     return parse_execution_errored;
 }
@@ -824,6 +821,7 @@ parse_execution_result_t parse_execution_context_t::populate_plain_process(
     bool expanded = expand_one(cmd, EXPAND_SKIP_CMDSUBST | EXPAND_SKIP_VARIABLES, NULL);
     if (!expanded) {
         report_error(statement, ILLEGAL_CMD_ERR_MSG, cmd.c_str());
+        proc_set_last_status(STATUS_ILLEGAL_CMD);
         return parse_execution_errored;
     }
 
@@ -878,7 +876,7 @@ parse_execution_result_t parse_execution_context_t::populate_plain_process(
         // If we have defined a wrapper around cd, use it, otherwise use the cd builtin.
         process_type = function_exists(L"cd") ? INTERNAL_FUNCTION : INTERNAL_BUILTIN;
     } else {
-        const globspec_t glob_behavior = contains(cmd, L"set", L"count") ? nullglob : failglob;
+        const globspec_t glob_behavior = (cmd == L"set" || cmd == L"count") ? nullglob : failglob;
         // Form the list of arguments. The command is the first argument. TODO: count hack, where we
         // treat 'count --help' as different from 'count $foo' that expands to 'count --help'. fish
         // 1.x never successfully did this, but it tried to!
@@ -945,16 +943,17 @@ parse_execution_result_t parse_execution_context_t::determine_arguments(
             case EXPAND_OK: {
                 break;
             }
+            default: {
+                DIE("unexpected expand_string() return value");
+                break;
+            }
         }
 
-        // Now copy over any expanded arguments. Do it using swap() to avoid extra allocations; this
+        // Now copy over any expanded arguments. Use std::move() to avoid extra allocations; this
         // is called very frequently.
-        size_t old_arg_count = out_arguments->size();
-        size_t new_arg_count = arg_expanded.size();
-        out_arguments->resize(old_arg_count + new_arg_count);
-        for (size_t i = 0; i < new_arg_count; i++) {
-            wcstring &new_arg = arg_expanded.at(i).completion;
-            out_arguments->at(old_arg_count + i).swap(new_arg);
+        out_arguments->reserve(out_arguments->size() + arg_expanded.size());
+        for (completion_t &new_arg : arg_expanded) {
+            out_arguments->push_back(std::move(new_arg.completion));
         }
     }
 
@@ -998,14 +997,12 @@ bool parse_execution_context_t::determine_io_chain(const parse_node_t &statement
                 if (target == L"-") {
                     new_io.reset(new io_close_t(source_fd));
                 } else {
-                    wchar_t *end = NULL;
-                    errno = 0;
-                    int old_fd = fish_wcstoi(target.c_str(), &end, 10);
-                    if (old_fd < 0 || errno || *end) {
-                        errored =
-                            report_error(redirect_node, _(L"Requested redirection to '%ls', which "
-                                                          L"is not a valid file descriptor"),
-                                         target.c_str());
+                    int old_fd = fish_wcstoi(target.c_str());
+                    if (errno || old_fd < 0) {
+                        const wchar_t *fmt =
+                            _(L"Requested redirection to '%ls', "
+                              L"which is not a valid file descriptor");
+                        errored = report_error(redirect_node, fmt, target.c_str());
                     } else {
                         new_io.reset(new io_fd_t(source_fd, old_fd, true));
                     }
@@ -1023,8 +1020,7 @@ bool parse_execution_context_t::determine_io_chain(const parse_node_t &statement
             }
             default: {
                 // Should be unreachable.
-                fprintf(stderr, "Unexpected redirection type %ld. aborting.\n",
-                        (long)redirect_type);
+                debug(0, "Unexpected redirection type %ld.", (long)redirect_type);
                 PARSER_DIE();
                 break;
             }
@@ -1037,7 +1033,7 @@ bool parse_execution_context_t::determine_io_chain(const parse_node_t &statement
     }
 
     if (out_chain && !errored) {
-        out_chain->swap(result);
+        *out_chain = std::move(result);
     }
     return !errored;
 }
@@ -1060,7 +1056,7 @@ parse_execution_result_t parse_execution_context_t::populate_boolean_process(
         }
         case parse_bool_not: {
             // NOT. Negate it.
-            job_set_flag(job, JOB_NEGATE, !job_get_flag(job, JOB_NEGATE));
+            job->set_flag(JOB_NEGATE, !job->get_flag(JOB_NEGATE));
             break;
         }
     }
@@ -1122,8 +1118,8 @@ parse_execution_result_t parse_execution_context_t::populate_job_process(
             break;
         }
         default: {
-            fprintf(stderr, "'%ls' not handled by new parser yet\n",
-                    specific_statement.describe().c_str());
+            debug(0, L"'%ls' not handled by new parser yet.",
+                  specific_statement.describe().c_str());
             PARSER_DIE();
             break;
         }
@@ -1148,9 +1144,9 @@ parse_execution_result_t parse_execution_context_t::populate_job_from_job_node(
     parse_execution_result_t result = parse_execution_success;
 
     // Create processes. Each one may fail.
-    std::vector<process_t *> processes;
-    processes.push_back(new process_t());
-    result = this->populate_job_process(j, processes.back(), *statement_node);
+    process_list_t processes;
+    processes.emplace_back(new process_t());
+    result = this->populate_job_process(j, processes.back().get(), *statement_node);
 
     // Construct process_ts for job continuations (pipelines), by walking the list until we hit the
     // terminal (empty) job continuation.
@@ -1173,29 +1169,23 @@ parse_execution_result_t parse_execution_context_t::populate_job_from_job_node(
         assert(statement_node != NULL);
 
         // Store the new process (and maybe with an error).
-        processes.push_back(new process_t());
-        result = this->populate_job_process(j, processes.back(), *statement_node);
+        processes.emplace_back(new process_t());
+        result = this->populate_job_process(j, processes.back().get(), *statement_node);
 
         // Get the next continuation.
         job_cont = get_child(*job_cont, 2, symbol_job_continuation);
         assert(job_cont != NULL);
     }
 
+    // Inform our processes of who is first and last
+    processes.front()->is_first_in_job = true;
+    processes.back()->is_last_in_job = true;
+
     // Return what happened.
     if (result == parse_execution_success) {
         // Link up the processes.
-        assert(!processes.empty());
-        j->first_process = processes.at(0);
-        for (size_t i = 1; i < processes.size(); i++) {
-            processes.at(i - 1)->next = processes.at(i);
-        }
-    } else {
-        // Clean up processes.
-        for (size_t i = 0; i < processes.size(); i++) {
-            const process_t *proc = processes.at(i);
-            processes.at(i) = NULL;
-            delete proc;
-        }
+        assert(!processes.empty());  //!OCLINT(multiple unary operator)
+        j->processes = std::move(processes);
     }
     return result;
 }
@@ -1208,12 +1198,10 @@ parse_execution_result_t parse_execution_context_t::run_1_job(const parse_node_t
 
     // Get terminal modes.
     struct termios tmodes = {};
-    if (shell_is_interactive()) {
-        if (tcgetattr(STDIN_FILENO, &tmodes)) {
-            // Need real error handling here.
-            wperror(L"tcgetattr");
-            return parse_execution_errored;
-        }
+    if (shell_is_interactive() && tcgetattr(STDIN_FILENO, &tmodes)) {
+        // Need real error handling here.
+        wperror(L"tcgetattr");
+        return parse_execution_errored;
     }
 
     // Increment the eval_level for the duration of this command.
@@ -1275,34 +1263,32 @@ parse_execution_result_t parse_execution_context_t::run_1_job(const parse_node_t
         return result;
     }
 
-    job_t *j = new job_t(acquire_job_id(), block_io);
-    j->tmodes = tmodes;
-    job_set_flag(j, JOB_CONTROL,
-                 (job_control_mode == JOB_CONTROL_ALL) ||
-                     ((job_control_mode == JOB_CONTROL_INTERACTIVE) && (shell_is_interactive())));
+    shared_ptr<job_t> job = std::make_shared<job_t>(acquire_job_id(), block_io);
+    job->tmodes = tmodes;
+    job->set_flag(JOB_CONTROL,
+                  (job_control_mode == JOB_CONTROL_ALL) ||
+                      ((job_control_mode == JOB_CONTROL_INTERACTIVE) && shell_is_interactive()));
 
-    job_set_flag(j, JOB_FOREGROUND, !tree.job_should_be_backgrounded(job_node));
+    job->set_flag(JOB_FOREGROUND, !tree.job_should_be_backgrounded(job_node));
 
-    job_set_flag(j, JOB_TERMINAL, job_get_flag(j, JOB_CONTROL) && !is_subshell && !is_event);
+    job->set_flag(JOB_TERMINAL, job->get_flag(JOB_CONTROL) && !is_event);
 
-    job_set_flag(j, JOB_SKIP_NOTIFICATION,
-                 is_subshell || is_block || is_event || !shell_is_interactive());
+    job->set_flag(JOB_SKIP_NOTIFICATION,
+                  is_subshell || is_block || is_event || !shell_is_interactive());
 
     // Tell the current block what its job is. This has to happen before we populate it (#1394).
-    parser->current_block()->job = j;
+    parser->current_block()->job = job;
 
     // Populate the job. This may fail for reasons like command_not_found. If this fails, an error
     // will have been printed.
     parse_execution_result_t pop_result =
-        this->populate_job_from_job_node(j, job_node, associated_block);
+        this->populate_job_from_job_node(job.get(), job_node, associated_block);
 
     // Clean up the job on failure or cancellation.
     bool populated_job = (pop_result == parse_execution_success);
     if (!populated_job || this->should_cancel_execution(associated_block)) {
-        assert(parser->current_block()->job == j);
+        assert(parser->current_block()->job == job);
         parser->current_block()->job = NULL;
-        delete j;
-        j = NULL;
         populated_job = false;
     }
 
@@ -1313,11 +1299,11 @@ parse_execution_result_t parse_execution_context_t::run_1_job(const parse_node_t
 
     if (populated_job) {
         // Success. Give the job to the parser - it will clean it up.
-        parser->job_add(j);
+        parser->job_add(job);
 
         // Check to see if this contained any external commands.
         bool job_contained_external_command = false;
-        for (const process_t *proc = j->first_process; proc != NULL; proc = proc->next) {
+        for (const auto &proc : job->processes) {
             if (proc->type == EXTERNAL) {
                 job_contained_external_command = true;
                 break;
@@ -1325,7 +1311,7 @@ parse_execution_result_t parse_execution_context_t::run_1_job(const parse_node_t
         }
 
         // Actually execute the job.
-        exec_job(*this->parser, j);
+        exec_job(*this->parser, job.get());
 
         // Only external commands require a new fishd barrier.
         if (job_contained_external_command) {
@@ -1338,7 +1324,7 @@ parse_execution_result_t parse_execution_context_t::run_1_job(const parse_node_t
         profile_item->level = eval_level;
         profile_item->parse = (int)(parse_time - start_time);
         profile_item->exec = (int)(exec_time - parse_time);
-        profile_item->cmd = j ? j->command() : wcstring();
+        profile_item->cmd = job ? job->command() : wcstring();
         profile_item->skipped = !populated_job;
     }
 
@@ -1370,7 +1356,7 @@ parse_execution_result_t parse_execution_context_t::run_job_list(const parse_nod
 parse_execution_result_t parse_execution_context_t::eval_node_at_offset(
     node_offset_t offset, const block_t *associated_block, const io_chain_t &io) {
     // Don't ever expect to have an empty tree if this is called.
-    assert(!tree.empty());
+    assert(!tree.empty());  //!OCLINT(multiple unary operator)
     assert(offset < tree.size());
 
     // Apply this block IO for the duration of this function.
@@ -1417,8 +1403,7 @@ parse_execution_result_t parse_execution_context_t::eval_node_at_offset(
         default: {
             // In principle, we could support other node types. However we never expect to be passed
             // them - see above.
-            fprintf(stderr, "Unexpected node %ls found in %s\n", node.describe().c_str(),
-                    __FUNCTION__);
+            debug(0, "Unexpected node %ls found in %s", node.describe().c_str(), __FUNCTION__);
             PARSER_DIE();
             break;
         }
