@@ -3,12 +3,15 @@
 // issues.
 #include "config.h"  // IWYU pragma: keep
 
-#include <assert.h>
 #include <errno.h>
+#include <stdio.h>
+#include <string.h>
 #include <sys/stat.h>
 #include <unistd.h>
 #include <wchar.h>
+
 #include <string>
+#include <type_traits>
 #include <vector>
 
 #include "common.h"
@@ -48,28 +51,29 @@ static bool path_get_path_core(const wcstring &cmd, wcstring *out_path,
     if (!bin_path_var.missing()) {
         bin_path = bin_path_var;
     } else {
-        if (contains(PREFIX L"/bin", L"/bin", L"/usr/bin")) {
-            bin_path = L"/bin" ARRAY_SEP_STR L"/usr/bin";
-        } else {
-            bin_path = L"/bin" ARRAY_SEP_STR L"/usr/bin" ARRAY_SEP_STR PREFIX L"/bin";
-        }
+        // Note that PREFIX is defined in the `Makefile` and is thus defined when this module is
+        // compiled. This ensures we always default to "/bin", "/usr/bin" and the bin dir defined
+        // for the fish programs. Possibly with a duplicate dir if PREFIX is empty, "/", "/usr" or
+        // "/usr/". If the PREFIX duplicates /bin or /usr/bin that is harmless other than a trivial
+        // amount of time testing a path we've already tested.
+        bin_path = L"/bin" ARRAY_SEP_STR L"/usr/bin" ARRAY_SEP_STR PREFIX L"/bin";
     }
 
-    wcstring nxt_path;
-    wcstokenizer tokenizer(bin_path, ARRAY_SEP_STR);
-    while (tokenizer.next(nxt_path)) {
-        if (nxt_path.empty()) continue;
-        append_path_component(nxt_path, cmd);
-        if (waccess(nxt_path, X_OK) == 0) {
+    std::vector<wcstring> pathsv;
+    tokenize_variable_array(bin_path, pathsv);
+    for (auto next_path : pathsv) {
+        if (next_path.empty()) continue;
+        append_path_component(next_path, cmd);
+        if (waccess(next_path, X_OK) == 0) {
             struct stat buff;
-            if (wstat(nxt_path, &buff) == -1) {
+            if (wstat(next_path, &buff) == -1) {
                 if (errno != EACCES) {
                     wperror(L"stat");
                 }
                 continue;
             }
             if (S_ISREG(buff.st_mode)) {
-                if (out_path) out_path->swap(nxt_path);
+                if (out_path) *out_path = std::move(next_path);
                 return true;
             }
             err = EACCES;
@@ -82,8 +86,9 @@ static bool path_get_path_core(const wcstring &cmd, wcstring *out_path,
                     break;
                 }
                 default: {
-                    debug(1, MISSING_COMMAND_ERR_MSG, nxt_path.c_str());
+                    debug(1, MISSING_COMMAND_ERR_MSG, next_path.c_str());
                     wperror(L"access");
+                    break;
                 }
             }
         }
@@ -124,24 +129,22 @@ bool path_get_cdpath(const wcstring &dir, wcstring *out, const wchar_t *wd,
         paths.push_back(path);
     } else {
         // Respect CDPATH.
-        env_var_t path = env_vars.get(L"CDPATH");
-        if (path.missing_or_empty()) path = L".";  // we'll change this to the wd if we have one
+        env_var_t cdpaths = env_vars.get(L"CDPATH");
+        if (cdpaths.missing_or_empty()) cdpaths = L".";
 
-        wcstring nxt_path;
-        wcstokenizer tokenizer(path, ARRAY_SEP_STR);
-        while (tokenizer.next(nxt_path)) {
-            if (nxt_path == L"." && wd != NULL) {
-                // nxt_path is just '.', and we have a working directory, so use the wd instead.
-                // TODO: if nxt_path starts with ./ we need to replace the . with the wd.
-                nxt_path = wd;
+        std::vector<wcstring> cdpathsv;
+        tokenize_variable_array(cdpaths, cdpathsv);
+        for (auto next_path : cdpathsv) {
+            if (next_path.empty()) next_path = L".";
+            if (next_path == L"." && wd != NULL) {
+                // next_path is just '.', and we have a working directory, so use the wd instead.
+                // TODO: if next_path starts with ./ we need to replace the . with the wd.
+                next_path = wd;
             }
-            expand_tilde(nxt_path);
+            expand_tilde(next_path);
+            if (next_path.empty()) continue;
 
-            // debug( 2, L"woot %ls\n", expanded_path.c_str() );
-
-            if (nxt_path.empty()) continue;
-
-            wcstring whole_path = nxt_path;
+            wcstring whole_path = next_path;
             append_path_component(whole_path, dir);
             paths.push_back(whole_path);
         }
@@ -189,19 +192,7 @@ wcstring path_apply_working_directory(const wcstring &path, const wcstring &work
 
     // We're going to make sure that if we want to prepend the wd, that the string has no leading
     // "/".
-    bool prepend_wd;
-    switch (path.at(0)) {
-        case L'/':
-        case HOME_DIRECTORY: {
-            prepend_wd = false;
-            break;
-        }
-        default: {
-            prepend_wd = true;
-            break;
-        }
-    }
-
+    bool prepend_wd = path.at(0) != L'/' && path.at(0) != HOME_DIRECTORY;
     if (!prepend_wd) {
         // No need to prepend the wd, so just return the path we were given.
         return path;
@@ -224,88 +215,100 @@ wcstring path_apply_working_directory(const wcstring &path, const wcstring &work
     return new_path;
 }
 
-static wcstring path_create_config() {
-    bool done = false;
-    wcstring res;
+/// We separate this from path_create() for two reasons. First it's only caused if there is a
+/// problem, and thus is not central to the behavior of that function. Second, we only want to issue
+/// the message once. If the current shell starts a new fish shell (e.g., by running `fish -c` from
+/// a function) we don't want that subshell to issue the same warnings.
+static void maybe_issue_path_warning(const wcstring &which_dir, const wcstring &custom_error_msg,
+                                     bool using_xdg, const wcstring &xdg_var, const wcstring &path,
+                                     int saved_errno) {
+    wcstring warning_var_name = L"_FISH_WARNED_" + which_dir;
+    if (env_exist(warning_var_name.c_str(), ENV_GLOBAL | ENV_EXPORT)) {
+        return;
+    }
+    env_set(warning_var_name, L"1", ENV_GLOBAL | ENV_EXPORT);
 
-    const env_var_t xdg_dir = env_get_string(L"XDG_CONFIG_HOME");
-    if (!xdg_dir.missing()) {
-        res = xdg_dir + L"/fish";
-        if (!create_directory(res)) {
-            done = true;
-        }
+    debug(0, custom_error_msg.c_str());
+    if (path.empty()) {
+        debug(0, _(L"Unable to locate the %ls directory."), which_dir.c_str());
+        debug(0, _(L"Please set the %ls or HOME environment variable before starting fish."),
+              xdg_var.c_str());
     } else {
-        const env_var_t home = env_get_string(L"HOME");
-        if (!home.missing()) {
-            res = home + L"/.config/fish";
-            if (!create_directory(res)) {
-                done = true;
-            }
-        }
+        const wchar_t *env_var = using_xdg ? xdg_var.c_str() : L"HOME";
+        debug(0, _(L"Unable to locate %ls directory derived from $%ls: '%ls'."), which_dir.c_str(),
+              env_var, path.c_str());
+        debug(0, _(L"The error was '%s'."), strerror(saved_errno));
+        debug(0, _(L"Please set $%ls to a directory where you have write access."), env_var);
     }
-
-    if (!done) {
-        res.clear();
-
-        debug(0, _(L"Unable to create a configuration directory for fish. Your personal settings "
-                   L"will not be saved. Please set the $XDG_CONFIG_HOME variable to a directory "
-                   L"where the current user has write access."));
-    }
-    return res;
+    fputwc(L'\n', stderr);
 }
 
-static wcstring path_create_data() {
-    bool done = false;
-    wcstring res;
+static void path_create(wcstring &path, const wcstring &xdg_var, const wcstring &which_dir,
+                        const wcstring &custom_error_msg) {
+    bool path_done = false;
+    bool using_xdg = false;
+    int saved_errno = 0;
 
-    const env_var_t xdg_dir = env_get_string(L"XDG_DATA_HOME");
-    if (!xdg_dir.missing()) {
-        res = xdg_dir + L"/fish";
-        if (!create_directory(res)) {
-            done = true;
+    // The vars we fetch must be exported. Allowing them to be universal doesn't make sense and
+    // allowing that creates a lock inversion that deadlocks the shell since we're called before
+    // uvars are available.
+    const env_var_t xdg_dir = env_get_string(xdg_var, ENV_GLOBAL | ENV_EXPORT);
+    if (!xdg_dir.missing_or_empty()) {
+        using_xdg = true;
+        path = xdg_dir + L"/fish";
+        if (create_directory(path) != -1) {
+            path_done = true;
+        } else {
+            saved_errno = errno;
         }
     } else {
-        const env_var_t home = env_get_string(L"HOME");
-        if (!home.missing()) {
-            res = home + L"/.local/share/fish";
-            if (!create_directory(res)) {
-                done = true;
+        const env_var_t home = env_get_string(L"HOME", ENV_GLOBAL | ENV_EXPORT);
+        if (!home.missing_or_empty()) {
+            path = home + (which_dir == L"config" ? L"/.config/fish" : L"/.local/share/fish");
+            if (create_directory(path) != -1) {
+                path_done = true;
+            } else {
+                saved_errno = errno;
             }
         }
     }
 
-    if (!done) {
-        res.clear();
-
-        debug(0, _(L"Unable to create a data directory for fish. Your history will not be saved. "
-                   L"Please set the $XDG_DATA_HOME variable to a directory where the current user "
-                   L"has write access."));
+    if (!path_done) {
+        maybe_issue_path_warning(which_dir, custom_error_msg, using_xdg, xdg_var, path,
+                                 saved_errno);
+        path.clear();
     }
-    return res;
+
+    return;
 }
 
 /// Cache the config path.
 bool path_get_config(wcstring &path) {
-    static const wcstring result = path_create_config();
-    path = result;
-    return !result.empty();
+    static bool config_path_done = false;
+    static wcstring config_path(L"");
+
+    if (!config_path_done) {
+        path_create(config_path, L"XDG_CONFIG_HOME", L"config",
+                    _(L"Your personal settings will not be saved."));
+        config_path_done = true;
+    }
+
+    path = config_path;
+    return !config_path.empty();
 }
 
 /// Cache the data path.
 bool path_get_data(wcstring &path) {
-    static const wcstring result = path_create_data();
-    path = result;
-    return !result.empty();
-}
+    static bool data_path_done = false;
+    static wcstring data_path(L"");
 
-__attribute__((unused)) static void replace_all(wcstring &str, const wchar_t *needle,
-                                                const wchar_t *replacement) {
-    size_t needle_len = wcslen(needle);
-    size_t offset = 0;
-    while ((offset = str.find(needle, offset)) != wcstring::npos) {
-        str.replace(offset, needle_len, replacement);
-        offset += needle_len;
+    if (!data_path_done) {
+        data_path_done = true;
+        path_create(data_path, L"XDG_DATA_HOME", L"data", _(L"Your history will not be saved."));
     }
+
+    path = data_path;
+    return !data_path.empty();
 }
 
 void path_make_canonical(wcstring &path) {
@@ -386,7 +389,7 @@ bool paths_are_same_file(const wcstring &path1, const wcstring &path2) {
     struct stat s1, s2;
     if (wstat(path1, &s1) == 0 && wstat(path2, &s2) == 0) {
         return s1.st_ino == s2.st_ino && s1.st_dev == s2.st_dev;
-    } else {
-        return false;
     }
+
+    return false;
 }
