@@ -2,7 +2,6 @@
 #define FISH_NO_ISW_WRAPPERS
 #include "config.h"
 
-#include <assert.h>
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -16,8 +15,9 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <wchar.h>
+#include <wctype.h>
+
 #include <map>
-#include <memory>
 #include <string>
 
 #include "common.h"
@@ -37,89 +37,99 @@ const file_id_t kInvalidFileID = {(dev_t)-1LL, (ino_t)-1LL, (uint64_t)-1LL, -1,
 #endif
 #endif
 
-/// Lock to protect wgettext.
-static pthread_mutex_t wgettext_lock;
-
 /// Map used as cache by wgettext.
-typedef std::map<wcstring, wcstring> wgettext_map_t;
-static wgettext_map_t wgettext_map;
+static owning_lock<std::map<wcstring, wcstring>> wgettext_map;
 
-bool wreaddir_resolving(DIR *dir, const std::wstring &dir_path, std::wstring &out_name,
-                        bool *out_is_dir) {
-    struct dirent *d = readdir(dir);
-    if (!d) return false;
-
-    out_name = str2wcstring(d->d_name);
-    if (out_is_dir) {
-        // The caller cares if this is a directory, so check.
-        bool is_dir = false;
-
-        // We may be able to skip stat, if the readdir can tell us the file type directly.
-        bool check_with_stat = true;
-#ifdef HAVE_STRUCT_DIRENT_D_TYPE
-        if (d->d_type == DT_DIR) {
-            // Known directory.
-            is_dir = true;
-            check_with_stat = false;
-        } else if (d->d_type == DT_LNK || d->d_type == DT_UNKNOWN) {
-            // We want to treat symlinks to directories as directories. Use stat to resolve it.
-            check_with_stat = true;
-        } else {
-            // Regular file.
-            is_dir = false;
-            check_with_stat = false;
-        }
-#endif  // HAVE_STRUCT_DIRENT_D_TYPE
-        if (check_with_stat) {
-            // We couldn't determine the file type from the dirent; check by stat'ing it.
-            cstring fullpath = wcs2string(dir_path);
-            fullpath.push_back('/');
-            fullpath.append(d->d_name);
-            struct stat buf;
-            if (stat(fullpath.c_str(), &buf) != 0) {
-                is_dir = false;
-            } else {
-                is_dir = !!(S_ISDIR(buf.st_mode));
-            }
-        }
-        *out_is_dir = is_dir;
+bool wreaddir_resolving(DIR *dir, const wcstring &dir_path, wcstring &out_name, bool *out_is_dir) {
+    struct dirent d;
+    struct dirent *result = NULL;
+    int retval = readdir_r(dir, &d, &result);
+    if (retval || !result) {
+        out_name = L"";
+        return false;
     }
+
+    out_name = str2wcstring(d.d_name);
+    if (!out_is_dir) return true;
+
+    // The caller cares if this is a directory, so check.
+    bool is_dir = false;
+    // We may be able to skip stat, if the readdir can tell us the file type directly.
+    bool check_with_stat = true;
+#ifdef HAVE_STRUCT_DIRENT_D_TYPE
+    if (d.d_type == DT_DIR) {
+        // Known directory.
+        is_dir = true;
+        check_with_stat = false;
+    } else if (d.d_type == DT_LNK || d.d_type == DT_UNKNOWN) {
+        // We want to treat symlinks to directories as directories. Use stat to resolve it.
+        check_with_stat = true;
+    } else {
+        // Regular file.
+        is_dir = false;
+        check_with_stat = false;
+    }
+#endif  // HAVE_STRUCT_DIRENT_D_TYPE
+    if (check_with_stat) {
+        // We couldn't determine the file type from the dirent; check by stat'ing it.
+        cstring fullpath = wcs2string(dir_path);
+        fullpath.push_back('/');
+        fullpath.append(d.d_name);
+        struct stat buf;
+        if (stat(fullpath.c_str(), &buf) != 0) {
+            is_dir = false;
+        } else {
+            is_dir = static_cast<bool>(S_ISDIR(buf.st_mode));
+        }
+    }
+    *out_is_dir = is_dir;
     return true;
 }
 
-bool wreaddir(DIR *dir, std::wstring &out_name) {
-    struct dirent *d = readdir(dir);
-    if (!d) return false;
+bool wreaddir(DIR *dir, wcstring &out_name) {
+    // We need to use a union to ensure that the dirent struct is large enough to avoid stomping on
+    // the stack. Some platforms incorrectly defined the `d_name[]` member as being one element
+    // long when it should be at least NAME_MAX + 1.
+    union {
+        struct dirent d;
+        char c[offsetof(struct dirent, d_name) + NAME_MAX + 1]; /* NAME_MAX is POSIX. */
+    } d_u;
+    struct dirent *result = NULL;
 
-    out_name = str2wcstring(d->d_name);
+    int retval = readdir_r(dir, &d_u.d, &result);
+    if (retval || !result) {
+        out_name = L"";
+        return false;
+    }
+
+    out_name = str2wcstring(d_u.d.d_name);
     return true;
 }
 
 bool wreaddir_for_dirs(DIR *dir, wcstring *out_name) {
+    struct dirent d;
     struct dirent *result = NULL;
-    while (result == NULL) {
-        struct dirent *d = readdir(dir);
-        if (!d) break;
+    while (!result) {
+        int retval = readdir_r(dir, &d, &result);
+        if (retval || !result) break;
 
 #if HAVE_STRUCT_DIRENT_D_TYPE
-        switch (d->d_type) {
+        switch (d.d_type) {
             case DT_DIR:
             case DT_LNK:
             case DT_UNKNOWN: {
-                // These may be directories.
-                result = d;
-                break;
+                break;  // these may be directories
             }
             default: {
-                // Nothing else can.
-                break;
+                break;  // nothing else can
             }
         }
 #else
         // We can't determine if it's a directory or not, so just return it.
-        result = d;
+        break;
 #endif
     }
+
     if (result && out_name) {
         *out_name = str2wcstring(result->d_name);
     }
@@ -186,11 +196,11 @@ bool set_cloexec(int fd) {
     int flags = fcntl(fd, F_GETFD, 0);
     if (flags < 0) {
         return false;
-    } else if (flags & FD_CLOEXEC) {
-        return true;
-    } else {
-        return fcntl(fd, F_SETFD, flags | FD_CLOEXEC) >= 0;
     }
+    if (flags & FD_CLOEXEC) {
+        return true;
+    }
+    return fcntl(fd, F_SETFD, flags | FD_CLOEXEC) >= 0;
 }
 
 static int wopen_internal(const wcstring &pathname, int flags, mode_t mode, bool cloexec) {
@@ -357,7 +367,7 @@ wchar_t *wrealpath(const wcstring &pathname, wchar_t *resolved_path) {
             if (pathsep_idx == cstring::npos) {
                 // No pathsep means a single path component relative to pwd.
                 narrow_res = realpath(".", NULL);
-                if (!narrow_res) DIE("unexpected realpath(\".\") failure");
+                assert(narrow_res != NULL);
                 pathsep_idx = 0;
             } else {
                 // Only call realpath() on the portion up to the last component.
@@ -405,7 +415,6 @@ wcstring wbasename(const wcstring &path) {
 
 // Really init wgettext.
 static void wgettext_really_init() {
-    pthread_mutex_init(&wgettext_lock, NULL);
     fish_bindtextdomain(PACKAGE_NAME, LOCALEDIR);
     fish_textdomain(PACKAGE_NAME);
 }
@@ -423,8 +432,8 @@ const wcstring &wgettext(const wchar_t *in) {
     wcstring key = in;
 
     wgettext_init_if_necessary();
-    scoped_lock locker(wgettext_lock);
-    wcstring &val = wgettext_map[key];
+    auto wmap = wgettext_map.acquire();
+    wcstring &val = wmap.value[key];
     if (val.empty()) {
         cstring mbs_in = wcs2string(key);
         char *out = fish_gettext(mbs_in.c_str());
@@ -464,6 +473,7 @@ int fish_iswalnum(wint_t wc) {
     return iswalnum(wc);
 }
 
+#if 0
 /// We need this because there are too many implementations that don't return the proper answer for
 /// some code points. See issue #3050.
 int fish_iswalpha(wint_t wc) {
@@ -471,6 +481,7 @@ int fish_iswalpha(wint_t wc) {
     if (fish_is_pua(wc)) return 0;
     return iswalpha(wc);
 }
+#endif
 
 /// We need this because there are too many implementations that don't return the proper answer for
 /// some code points. See issue #3050.
@@ -480,34 +491,6 @@ int fish_iswgraph(wint_t wc) {
     return iswgraph(wc);
 }
 
-/// Test if the given string is a valid variable name.
-///
-/// \return null if this is a valid name, and a pointer to the first invalid character otherwise.
-const wchar_t *wcsvarname(const wchar_t *str) {
-    while (*str) {
-        if ((!fish_iswalnum(*str)) && (*str != L'_')) {
-            return str;
-        }
-        str++;
-    }
-    return NULL;
-}
-
-/// Test if the given string is a valid variable name.
-///
-/// \return null if this is a valid name, and a pointer to the first invalid character otherwise.
-const wchar_t *wcsvarname(const wcstring &str) { return wcsvarname(str.c_str()); }
-
-/// Test if the given string is a valid function name.
-///
-/// \return null if this is a valid name, and a pointer to the first invalid character otherwise.
-const wchar_t *wcsfuncname(const wcstring &str) { return wcschr(str.c_str(), L'/'); }
-
-/// Test if the given string is valid in a variable name.
-///
-/// \return true if this is a valid name, false otherwise.
-bool wcsvarchr(wchar_t chr) { return fish_iswalnum(chr) || chr == L'_'; }
-
 /// Convenience variants on fish_wcwswidth().
 ///
 /// See fallback.h for the normal definitions.
@@ -518,16 +501,138 @@ int fish_wcswidth(const wchar_t *str) { return fish_wcswidth(str, wcslen(str));
 /// See fallback.h for the normal definitions.
 int fish_wcswidth(const wcstring &str) { return fish_wcswidth(str.c_str(), str.size()); }
 
-int fish_wcstoi(const wchar_t *str, wchar_t **endptr, int base) {
-    long ret = wcstol(str, endptr, base);
-    if (ret > INT_MAX) {
-        ret = INT_MAX;
+/// Like fish_wcstol(), but fails on a value outside the range of an int.
+///
+/// This is needed because BSD and GNU implementations differ in several ways that make it really
+/// annoying to use them in a portable fashion.
+///
+/// The caller doesn't have to zero errno. Sets errno to -1 if the int ends with something other
+/// than a digit. Leading whitespace is ignored (per the base wcstol implementation). Trailing
+/// whitespace is also ignored. We also treat empty strings and strings containing only whitespace
+/// as invalid.
+int fish_wcstoi(const wchar_t *str, const wchar_t **endptr, int base) {
+    while (iswspace(*str)) ++str;  // skip leading whitespace
+    if (!*str) {  // this is because some implementations don't handle this sensibly
+        errno = EINVAL;
+        if (endptr) *endptr = str;
+        return 0;
+    }
+
+    errno = 0;
+    wchar_t *_endptr;
+    long result = wcstol(str, &_endptr, base);
+    if (result > INT_MAX) {
+        result = INT_MAX;
         errno = ERANGE;
-    } else if (ret < INT_MIN) {
-        ret = INT_MIN;
+    } else if (result < INT_MIN) {
+        result = INT_MIN;
         errno = ERANGE;
     }
-    return (int)ret;
+    while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
+    if (!errno && *_endptr) {
+        if (_endptr == str) {
+            errno = EINVAL;
+        } else {
+            errno = -1;
+        }
+    }
+    if (endptr) *endptr = _endptr;
+    return (int)result;
+}
+
+/// An enhanced version of wcstol().
+///
+/// This is needed because BSD and GNU implementations differ in several ways that make it really
+/// annoying to use them in a portable fashion.
+///
+/// The caller doesn't have to zero errno. Sets errno to -1 if the int ends with something other
+/// than a digit. Leading whitespace is ignored (per the base wcstol implementation). Trailing
+/// whitespace is also ignored.
+long fish_wcstol(const wchar_t *str, const wchar_t **endptr, int base) {
+    while (iswspace(*str)) ++str;  // skip leading whitespace
+    if (!*str) {  // this is because some implementations don't handle this sensibly
+        errno = EINVAL;
+        if (endptr) *endptr = str;
+        return 0;
+    }
+
+    errno = 0;
+    wchar_t *_endptr;
+    long result = wcstol(str, &_endptr, base);
+    while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
+    if (!errno && *_endptr) {
+        if (_endptr == str) {
+            errno = EINVAL;
+        } else {
+            errno = -1;
+        }
+    }
+    if (endptr) *endptr = _endptr;
+    return result;
+}
+
+/// An enhanced version of wcstoll().
+///
+/// This is needed because BSD and GNU implementations differ in several ways that make it really
+/// annoying to use them in a portable fashion.
+///
+/// The caller doesn't have to zero errno. Sets errno to -1 if the int ends with something other
+/// than a digit. Leading whitespace is ignored (per the base wcstoll implementation). Trailing
+/// whitespace is also ignored.
+long long fish_wcstoll(const wchar_t *str, const wchar_t **endptr, int base) {
+    while (iswspace(*str)) ++str;  // skip leading whitespace
+    if (!*str) {  // this is because some implementations don't handle this sensibly
+        errno = EINVAL;
+        if (endptr) *endptr = str;
+        return 0;
+    }
+
+    errno = 0;
+    wchar_t *_endptr;
+    long long result = wcstoll(str, &_endptr, base);
+    while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
+    if (!errno && *_endptr) {
+        if (_endptr == str) {
+            errno = EINVAL;
+        } else {
+            errno = -1;
+        }
+    }
+    if (endptr) *endptr = _endptr;
+    return result;
+}
+
+/// An enhanced version of wcstoull().
+///
+/// This is needed because BSD and GNU implementations differ in several ways that make it really
+/// annoying to use them in a portable fashion.
+///
+/// The caller doesn't have to zero errno. Sets errno to -1 if the int ends with something other
+/// than a digit. Leading minus is considered invalid. Leading whitespace is ignored (per the base
+/// wcstoull implementation). Trailing whitespace is also ignored.
+unsigned long long fish_wcstoull(const wchar_t *str, const wchar_t **endptr, int base) {
+    while (iswspace(*str)) ++str;  // skip leading whitespace
+    if (!*str ||      // this is because some implementations don't handle this sensibly
+        *str == '-')  // disallow minus as the first character to avoid questionable wrap-around
+    {
+        errno = EINVAL;
+        if (endptr) *endptr = str;
+        return 0;
+    }
+
+    errno = 0;
+    wchar_t *_endptr;
+    unsigned long long result = wcstoull(str, &_endptr, base);
+    while (iswspace(*_endptr)) ++_endptr;  // skip trailing whitespace
+    if (!errno && *_endptr) {
+        if (_endptr == str) {
+            errno = EINVAL;
+        } else {
+            errno = -1;
+        }
+    }
+    if (endptr) *endptr = _endptr;
+    return result;
 }
 
 file_id_t file_id_t::file_id_from_stat(const struct stat *buf) {
@@ -560,7 +665,7 @@ file_id_t file_id_t::file_id_from_stat(const struct stat *buf) {
 file_id_t file_id_for_fd(int fd) {
     file_id_t result = kInvalidFileID;
     struct stat buf = {};
-    if (0 == fstat(fd, &buf)) {
+    if (fd >= 0 && 0 == fstat(fd, &buf)) {
         result = file_id_t::file_id_from_stat(&buf);
     }
     return result;
